/*
 * Copyright (C) 2015 Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file Detumble.c
 * @author Brendan Bruner
 * @date 2014-12-28
 */

#ifndef COMMAND_H_
#define COMMAND_H_

#include "driver_toolkit.h"
#include <stdint.h>
#include "lpc_types.h"
#include "LPC17xx.h"

/************************************************************************/
/* Possible execution statuses											*/
/************************************************************************/
typedef enum
{
	EXECUTING = 0,
	EXECUTED,
	PENDING
} command_status_t;

/************************************************************************/
/* Typedefs required for command_t structure							*/
/************************************************************************/
typedef struct command_t command_t;
typedef void (*execution_handler)( command_t * );
typedef struct system_state_relay_t system_state_relay_t;

/************************************************************************/
/* command_t Structure Prototype										*/
/************************************************************************/
/**
 * @struct command_t
 * @brief
 * 		The generic structure of a command.
 * @details
 * 		The generic structure of a command. All instances of this structure
 * 		are initialized by a command specific initialization method. After
 * 		initialization, the command can then be used to perform a specific
 * 		action. The execution of commands will almost always be relayed
 * 		to an instance of the system_state_relay_t structure. This relay
 * 		structure will perform state specific selection of commands to
 * 		execute.
 *
 * 		A command_t structure must remain valid in memory until it is
 * 		finished execution. Whether it is in .bss, stack, or heap does
 * 		not matter.
 *
 * 		For example,
 * 			@code
 * 			command_t reset_power;
 *
 * 			initialize_command_reset_power( &reset_power, NULL );
 * 			relay_execute_high_priority_command( &relay, &reset_power );
 * 			@endcode
 * 		In this example, the variable 'relay' is of type system_state_relay_t
 * 		and is assumed to have been initialized elsewhere in code.
 *
 * @var command_t::generic_data
 * 		A pointer to any generic data. This data gets set at the time a
 * 		command is initialized. The implementation of the commands
 * 		execute method can then make use of this data (if needed).
 * 		For example,
 * 			@code
 * 			command_t 	inc_counter;
 * 			int			count;
 *
 * 			count = 0;
 * 			initialize_command_inc_counter( &inc_counter, &count );
 *
 * 			inc_counter.execute( &inc_counter );
 * 			@endcode
 * 		In this example, the command is assumed to increment an integer variable.
 * 		Note, directly calling the execute method is uncommon.
 *
 * @var command_t::execute
 * 		A method pointer, it is assigned based on the command initialization
 * 		method called:
 * 			@code
 * 			void execute( command_t * );
 * 			@endcode
 * 		Note, this method takes a pointer to itself. This means, the command_t
 * 		structure where this method is called from is passed in as an argument
 * 		to the method.
 * 			@code
 * 			example_command.execute( &example_command );
 * 			@endcode
 * 		In this example, the variable 'example_command' is assumed to be of type
 * 		command_t and has been initialized properly.
 *
 * 		Note, this method should typically not be called directly. Instead,
 * 		call the global function:
 * 			@code
 * 			void execute_command( comand_t * );
 * 			@endcode
 * 		For example:
 * 			@code
 * 			command_t command;
 *
 * 			initialize_command( &command );
 * 			//Do this
 * 			execute_command( &command );
 * 			//Not this
 * 			command.execute( &command );
 * 			@endcode
 *
 * 		The global function should be called because every command
 * 		requires some generic pre-processing to be done before their unique
 * 		command_t::execute method is called. The global method performs this
 * 		pre-processing then invokes the command_t::execute method.
 */
struct command_t
{
	void 				*generic_data;
	command_status_t	__execution_status;
	execution_handler 	execute;
};

/************************************************************************/
/* Public Methods														*/
/************************************************************************/
/**
 * @brief
 * 		Call this method to execute a command.
 *
 * @details
 * 		Call this method to execute a command.
 *
 * @param command
 * 		A pointer to the command_t structure to execute.
 *
 * @memberof command_t
 */
void execute_command( command_t *command );

/**
 * @brief
 * 		Returns the current execution status of a command.
 *
 * @details
 * 		Returns the current execution status of a command.
 *
 * @param command
 * 		A pointer to the command_t structure to execute.
 *
 * @returns
 * 		EXECUTING		If the command is currently  being executed.
 * 		<br>EXECUTED	If the command has been executed.
 * 		<br>PENDING		If the command is waiting to be executed.
 *
 * @memberof command_t
 */
command_status_t command_status( command_t *command );


/************************************************************************/
/* Collect Housekeeping Commands										*/
/************************************************************************/

/**
 * @brief
 * 		Initializes a command_t structure to invoke the collection of
 * 		hk from the EPS
 * @details
 * 		Initializes a command_t structure to invoke the collection of
 * 		hk from the EPS
 * @param command
 * 		The command_t structure to initialize.
 * @param kit
 * 		A pointer to the driver_toolkit_t structure to collect hk from.
 * @memberof command_t
 */
void initialize_command_collect_eps_hk(	command_t *command,
										driver_toolkit_t *kit );

/**
 * @brief
 * 		Initializes a command_t structure to invoke the collection of
 * 		hk from the ADCS
 * @details
 * 		Initializes a command_t structure to invoke the collection of
 * 		hk from the ADCS
 * @param command
 * 		The command_t structure to initialize.
 * @param kit
 * 		A pointer to the driver_toolkit_t structure to collect hk from.
 * @memberof command_t
 */
void initialize_command_collect_adcs_hk( command_t *command,
										 driver_toolkit_t *kit );

/**
 * @brief
 * 		Initializes a command_t structure to invoke the collection of
 * 		hk from the COMM
 * @details
 * 		Initializes a command_t structure to invoke the collection of
 * 		hk from the COMM
 * @param command
 * 		The command_t structure to initialize.
 * @param kit
 * 		A pointer to the driver_toolkit_t structure to collect hk from.
 * @memberof command_t
 */
void initialize_command_collect_comm_hk( command_t *command,
										 driver_toolkit_t *kit );


/************************************************************************/
/* Set State Commands													*/
/************************************************************************/
/**
 * @brief
 * 		Initializes a command_t structure to change the systems state
 * 		to power safe.
 * @details
 * 		Initializes a command_t structure to change the systems state
 * 		to power safe. The second argument to the method (the data
 * 		parameter) MUST be a pointer to a system_struct_relay_t. The
 * 		state will be changed in the system_struct_relay_t passed in.
 * @param command
 * 		The command_t structure to initialize.
 * @param data
 * 		Must be a pointer to the system_struct_relay_t variable that
 * 		should have its current state changed to the power safe state.
 * @memberof command_t
 */
void initialize_command_set_state_power_safe( 	command_t *command,
												system_state_relay_t *data );


/************************************************************************/
/* Transmit Data Commands												*/
/************************************************************************/
/**
 * @brief
 * 		Initialize a command_t structure to transmit eps housekeeping
 * 		to ground station.
 *
 * @details
 * 		Initialize a command_t structure to transmit eps housekeeping
 * 		to ground station.
 *
 * 		The command begins by extracting eps housekeeping from the sd
 * 		card and transmitting the data to ground station. The transmitted
 * 		data is then written to a log file on the sd card and deleted
 * 		from the file it was extracted from. This serves as a back up
 * 		in case transmitted data is lost.
 *
 * 		This process continues until a hardwired number of bytes has
 * 		been transmitted.
 *
 * @param command
 * 		A pointer to the command_t structure to initialize.
 *
 * @param kit
 * 		A pointer to the driver_toolkit_t structure the command will use
 * 		for talking to ground station and extracting data from sd card.
 */
void initialize_command_transmit_eps_hk(	command_t *command,
											driver_toolkit_t *kit );

/**
 * @brief
 * 		Initialize a command_t structure to transmit adcs housekeeping
 * 		to ground station.
 *
 * @details
 * 		Initialize a command_t structure to transmit adcs housekeeping
 * 		to ground station.
 *
 * 		The command begins by extracting adcs housekeeping from the sd
 * 		card and transmitting the data to ground station. The transmitted
 * 		data is then written to a log file on the sd card and deleted
 * 		from the file it was extracted from. This serves as a back up
 * 		in case transmitted data is lost.
 *
 * 		This process continues until a hardwired number of bytes has
 * 		been transmitted.
 *
 * @param command
 * 		A pointer to the command_t structure to initialize.
 *
 * @param kit
 * 		A pointer to the driver_toolkit_t structure the command will use
 * 		for talking to ground station and extracting data from sd card.
 */
void initialize_command_transmit_adcs_hk(	command_t *command,
											driver_toolkit_t *kit );

/**
 * @brief
 * 		Initialize a command_t structure to transmit comm housekeeping
 * 		to ground station.
 *
 * @details
 * 		Initialize a command_t structure to transmit comm housekeeping
 * 		to ground station.
 *
 * 		The command begins by extracting comm housekeeping from the sd
 * 		card and transmitting the data to ground station. The transmitted
 * 		data is then written to a log file on the sd card and deleted
 * 		from the file it was extracted from. This serves as a back up
 * 		in case transmitted data is lost.
 *
 * 		This process continues until a hard wired number of bytes has been
 * 		transmitted.
 *
 * @param command
 * 		A pointer to the command_t structure to initialize.
 *
 * @param kit
 * 		A pointer to the driver_toolkit_t structure the command will use
 * 		for talking to ground station and extracting data from sd card.
 */
void initialize_command_transmit_comm_hk(	command_t *command,
											driver_toolkit_t *kit );

void initialize_command_transmit_state_hk(	command_t *command,
											driver_toolkit_t *kit );
/************************************************************************/
/* Commands for Testing Purposes										*/
/************************************************************************/
void initialize_command_print_string( command_t *, char const * );
void initialize_command_receive_time_from_adcs_over_i2c( command_t *self, driver_toolkit_t *drivers );
void initialize_command_send_time_to_adcs_over_i2c( command_t *self, driver_toolkit_t *drivers );
void initialize_command_print_state( command_t *, system_state_relay_t * );

#endif /* COMMAND_H_ */
