/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file system_state_relay
 * @author Brendan Bruner
 * @date 2015-01-28
 */
#ifndef INCLUDE_SYSTEM_STATE_RELAY_H_
#define INCLUDE_SYSTEM_STATE_RELAY_H_

#include "system_states.h"
#include "commands.h"
#include "driver_toolkit.h"
#include "FreeRTOS.h"

/**
 * @struct system_state_relay_t
 * @brief
 * 		A structure for relaying information to the systems current state.
 *
 * @details
 * 		A structure for relaying information to the systems current state.
 * 		All access to state behavior is made through this structure. It takes
 * 		commands (command_t structures), which perform state specific actions, and
 * 		relays that command_t to the current state of the system. The current
 * 		state of the system is represented by a system_state_t structure.
 *
 * 		For example, lets say we have command_t which will turn on a power
 * 		hungry subsystem to briefly do IO with it.
 * 			@code
 * 			command_t command;
 *
 * 			initialize_power_hungry_command( &command, NULL );
 * 			relay_command_very_high_priority( &relay, &command );
 * 			@endcode
 * 		Assume the variable 'relay' is of type system_state_relay_t and
 * 		has been initialized somewhere else. In this example, the power
 * 		hungry command will be executed based on what the current state is.
 * 		If the current state is power safe then the command will not be
 * 		executed, but if the current state is the science state then the
 * 		command will be executed.
 *
 * @attention
 * 		DO NOT write to any structure members directly. Use provided API
 * 		for writes. Some members can be read from, others cannot. Read
 * 		documentation on each variable for more information.
 *
 * @var system_state_relay_t::current_state
 * 		This variable should NEVER be accessed directly. For example,
 * 			@code
 * 			system_state_relay_t relay;
 *
 * 			initialize_system_state_relay( &relay );
 * 			relay.current_state = SOME_VALUE;
 * 			@endcode
 * 		This should NEVER be done.
 * 		<br>This variable is protected by a mutex.
 *
 * @var system_state_relay_t::power_on
 * 		An instance of a system_state_t structure initialized to behave
 * 		as a a power on state. This variable should never be written to.
 *
 * @var system_state_relay_t::post_ejection
 * 		An instance of a system_state_t structure initialized to behave
 * 		as a a post ejection state. This variable should never be written to.
 *
 * @var system_state_relay_t::pre_charge
 * 		An instance of a system_state_t structure initialized to behave
 * 		as a a pre charge state. This variable should never be written to.
 *
 * @var system_state_relay_t::start_up
 * 		An instance of a system_state_t structure initialized to behave
 * 		as a a start up state. This variable should never be written to.
 *
 * @var system_state_relay_t::power_safe
 * 		An instance of a system_state_t structure initialized to behave
 * 		as a a power safe state. This variable should never be written to.
 *
 * @var system_state_relay_t::detumble
 * 		An instance of a system_state_t structure initialized to behave
 * 		as a a detumble state. This variable should never be written to.
 *
 * @var system_state_relay_t::science
 * 		An instance of a system_state_t structure initialized to behave
 * 		as a a science state. This variable should never be written to.
 *
 * @var system_state_relay_t::transmit
 * 		An instance of a system_state_t structure initialized to behave
 * 		as a a transmit state. This variable should never be written to.
 *
 * @var system_state_relay_t::drivers
 * 		A pointer to a driver_toolkit_t structure, This variable should never
 * 		be written to, but can be read from. The system_states being managed
 * 		by the relay will use these drivers as part of their next state logic and
 * 		possibly for other functionality.
 */
struct system_state_relay_t
{
	system_state_t 	*current_state;
	system_state_t 	power_on,
					post_ejection,
					pre_charge,
					start_up,
					power_safe,
					detumble,
					science,
					transmit;
	driver_toolkit_t	*drivers;
	SemaphoreHandle_t	__state_mutex;
};

/**
 * @brief
 * 		Initializes a system_state_relay_t structure.
 *
 * @details
 * 		Initializes a system_state_relay_t structure. All instances
 * 		of system_state_relay_t are useless until initialize and will
 * 		lead to undefined behavior if not initialized.
 *
 * @param relay
 * 		A pointer to the system_state_relay_t structure to initialize.
 *
 * @param initial_state
 * 		A pointer to a system_state_t structure. This is the initial
 * 		state of the system and its entry method will be called. If
 * 		no initial state is specified (ie, NULL is passed in) then
 * 		the initial state of the system will be set to the power on
 * 		state.
 *
 * 		If an initial state is specified then it MUST be initialized
 * 		before being given used in this method.
 *
 * @param drivers
 * 		A pointer to the driver_toolkit_t structure which will be used.
 * 		This is the driver toolkit which will be made available to all
 * 		the system_states.
 *
 * @memberof system_state_relay_t
 */
void initialize_system_state_relay( system_state_relay_t *relay,
									system_state_t *initial_state,
									driver_toolkit_t *drivers );

/**
 * @brief
 * 		Changes the current state of the system.
 *
 * @details
 * 		Changes the current state of the system. The current states exit()
 * 		method is called. The state is changed, then the new states entry()
 * 		method is called.
 *
 * @attention
 * 		Using this method will likely have unintended side effects. Only
 * 		use this method if you know what you're doing.
 *
 * @param relay
 * 		A pointer to the system_state_relay_t structure to change
 * 		the state of.
 *
 * @param next
 * 		A pointer to the system_state_t structure which will be the new
 * 		state. Typically, this pointer will be a member of the system_state_relay_t
 * 		structure passed in as argument 1. For example,
 * 			@code
 * 			relay_change_system_state( &relay, &relay.power_safe );
 * 			@endcode
 * 		However, if this is not the case, then the system_state_t structure
 * 		must remain valid in memory for at least the duration it is the current
 * 		state.
 *
 * 		If NULL is passed in as the next state the function returns immediately
 * 		without doing anything.
 *
 * @memberof system_state_relay_t
 */
void relay_force_change_system_state(	system_state_relay_t *relay,
										system_state_t *next );

/**
 * @brief
 * 		Relays a very high priority command to the current state.
 *
 * @details
 * 		Relays a very high priority command to the current state.
 *
 * @param relay
 *		A pointer to the system_state_relay_t structure which will
 *		be used to relay the command.
 *
 * @param command
 * 		The command to relay to the current state. If this is NULL
 * 		the function immediately returns.
 *
 * @memberof system_state_relay_t
 */
void relay_command_very_high_priority( 	system_state_relay_t *relay,
										command_t *command );

/**
 * @brief
 * 		Relays a high priority command to the current state.
 *
 * @details
 * 		Relays a high priority command to the current state.
 *
 * @param relay
 *		A pointer to the system_state_relay_t structure which will
 *		be used to relay the command.
 *
 * @param command
 * 		The command to relay to the current state. If this is NULL
 * 		the function immediately returns.
 *
 * @memberof system_state_relay_t
 */
void relay_command_high_priority( 	system_state_relay_t *relay,
									command_t *command );

/**
 * @brief
 * 		Relays a low priority command to the current state.
 *
 * @details
 * 		Relays a low priority command to the current state.
 *
 * @param relay
 *		A pointer to the system_state_relay_t structure which will
 *		be used to relay the command.
 *
 * @param command
 * 		The command to relay to the current state. If this is NULL
 * 		the function immediately returns.
 *
 * @memberof system_state_relay_t
 */
void relay_command_low_priority(	system_state_relay_t *relay,
									command_t *command );

/**
 * @brief
 * 		Relays a very low priority command to the current state.
 *
 * @details
 * 		Relays a very low priority command to the current state.
 *
 * @param relay
 *		A pointer to the system_state_relay_t structure which will
 *		be used to relay the command.
 *
 * @param command
 * 		The command to relay to the current state. If this is NULL
 * 		the function immediately returns.
 *
 * @memberof system_state_relay_t
 */
void relay_command_very_low_priority(	system_state_relay_t *relay,
										command_t *command );

/**
 * @brief
 * 		Relays a command to collect housekeeping data to the current state.
 *
 * @details
 * 		Relays a command to collect housekeeping data to the current state.
 *
 * @param relay
 *		A pointer to the system_state_relay_t structure which will
 *		be used to relay the command.
 *
 * @param command
 * 		The command which will collect housekeeping data. This will be
 * 		relayed to the current state. If this is NULL the function
 * 		immediately returns.
 *
 * @memberof system_state_relay_t
 */
void relay_command_collect_housekeeping(	system_state_relay_t *relay,
											command_t *command );

/**
 * @brief
 * 		Relays a command to transmit data to the current state.
 *
 * @details
 * 		Relays a command which will transmit data to a ground station.
 * 		This command is relayed to the current state.
 *
 * @param relay
 *		A pointer to the system_state_relay_t structure which will
 *		be used to relay the command.
 *
 * @param command
 * 		A transmit data command to relay to the current state. If this is NULL
 * 		the function immediately returns.
 *
 * @memberof system_state_relay_t
 */
void relay_command_transmit_data(	system_state_relay_t *relay,
									command_t *command );

/**
 * @brief
 * 		Relays commands which handle the MnLP to the current state.
 *
 * @details
 *		Relays commands which handle the MnLP to the current state.
 *
 * @param relay
 *		A pointer to the system_state_relay_t structure which will
 *		be used to relay the command.
 *
 * @param command
 * 		The command which will handle the MnLP. This command is relayed
 * 		to the current state. If this is NULL the function immediately
 * 		returns.
 *
 * @memberof system_state_relay_t
 */
void relay_command_handle_MnLP(	system_state_relay_t *relay,
								command_t *command );

/**
 * @brief
 * 		Relays commands which handle the DFGM to the current state.
 *
 * @details
 * 		Relays commands which handle the DFGM to the current state.
 *
 * @param relay
 *		A pointer to the system_state_relay_t structure which will
 *		be used to relay the command.
 *
 * @param command
 * 		The command which will handle the DFGM. This command is relayed
 * 		to the current state. If this is NULL the function immediately
 * 		returns.
 *
 * @memberof system_state_relay_t
 */
void relay_command_handle_DFGM(	system_state_relay_t *relay,
								command_t *command );

/**
 * @brief
 * 		Gets the integer id of the current state.
 *
 * @details
 *		Gets the integer id of the current state.
 *
 *@param relay
 *		A pointer to the system_state_relay_t structure who's current state
 *		id is of interest.
 *
 * @returns
 * 		The id of the current state.
 *
 * @memberof system_state_relay_t
 */
system_state_id_t relay_system_state_id( system_state_relay_t *relay );

#endif /* INCLUDE_SYSTEM_STATE_RELAY_H_ */
