/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file system_states
 * @author Brendan Bruner
 * @date 
 *
 * Defines the common structure used  by states and prototypes the
 * methods to initialize this structure depending on the state.
 *
 * All state ids are #defined here.
 */
#ifndef INCLUDE_SYSTEM_STATES_H_
#define INCLUDE_SYSTEM_STATES_H_

#include "commands.h"

#ifndef NULL
#define NULL ((void *) 0)
#endif

/* Forward declaration for system_state_relay_t and system_state_t. */
typedef struct system_state_relay_t system_state_relay_t;
typedef struct system_state_t		system_state_t;

typedef unsigned short system_state_id_t;
typedef void (*command_handler)( system_state_t *, command_t * );
typedef void (*entry_handler)( system_state_t *, system_state_relay_t * );
typedef system_state_t *(*next_state_handler)( 	system_state_t *,
												system_state_relay_t * );

#define STATE_POWER_ON_ID 		((system_state_id_t) 0)
#define STATE_POST_EJECTION_ID	((system_state_id_t) 1)
#define STATE_PRE_CHARGE_ID		((system_state_id_t) 2)
#define STATE_START_UP_ID		((system_state_id_t) 3)
#define STATE_POWER_SAFE_ID		((system_state_id_t) 4)
#define STATE_SCIENCE_ID		((system_state_id_t) 5)
#define STATE_DETUMBLE_ID		((system_state_id_t) 6)
#define STATE_TRANSMIT_ID		((system_state_id_t) 7)
#define STATE_NULL				((system_state_id_t) 8)
extern char const * const state_id_to_name[]; /* declare */

/**
 * @struct system_state_t
 * @brief
 * 		The generic structure of a system state.
 *
 * @details
 * 		The generic structure of a system state. All instances of
 * 		this structure are initialized by a state specific initialization
 * 		method and can then be used for state specific behavior.
 * 		All instances of this structure must remain valid in memory
 * 		whether they are in .bss, heap, or stack space.
 *
 * 		For example,
 * 			@code
 * 			system_state_t power_on, science;
 *
 * 			initialize_power_on_state( &power_on );
 * 			initialize_science_state( &science );
 * 			@endcode
 * 		The variables 'science' and 'power_on' will provide different
 * 		behavior.
 * 			@code
 * 			power_on.handle_MnLP( &power_on, &some_command );
 * 			science.handle_MnLP( &power_on, &some_command );
 * 			@endcode
 * 		The two handlers will very likely do something different. This
 * 		is the state specific behaviour.
 *
 * @remark
 * 		Source files which use this structure must include the following header
 * 			@code
 * 			#include "system_state_relay.h"
 * 			@endcode
 *
 * @var system_state_t::generic_data
 * 		A pointer to and data a state might need. Some state implementations
 * 		may needs this, some may not.
 *
 * @var system_state_t::id
 * 		A unique integer number which represents the state. This
 * 		is useful for querying the current state of the system
 * 		or for logging state changes.
 * 		After initialization, this variable should never be written to.
 *
 * @var system_state_t::enter_system_state
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void enter_system_state( system_state_t *, system_state_relay_t * );
 * 			@endcode
 * 		The method is called before a transition INTO the state is made
 * 		and should only be called within system_state_relay_t specific methods.
 * 		This enables set up code to be executed before any state specific methods
 * 		are called. The input to the method must be the address of the
 * 		structure being used to index the method and the system_state_relay_t
 * 		invoking the method.
 * 			@code
 * 			var.enter_system_state( &var, &relay );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * @var system_state_t::exit_system_state
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void exit_system_state( system_state_t *, system_state_relay_t * );
 * 			@endcode
 * 		The method is called before a transition OUT OF the state is made
 * 		and should only be called within system_state_relay_t specific methods.
 * 		This enables tear down code to be executed before the state is left.
 * 		The input to the method must be the address of the structure
 * 		being used to index the method and the system_state_relay_t invoking
 * 		the method.
 * 			@code
 * 			var.exit_system_state( &var, &relay );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * @var system_state_t::next_system_state
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			system_state_t *next_system_state( 	system_state_t *,
 * 												system_state_relay_t * );
 * 			@endcode
 * 		This method defines next state logic. It returns a pointer
 * 		to the next state, but does not make the transition. It
 * 		uses the system_state_relay_t structure passed in to find the
 * 		pointer to the next state. The system_state_t pointer must be the
 * 		address of the structure being used to index the method.
 * 			@code
 * 			var.next_system_state( &var, &relay );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * 		The method will return NULL if the system does not need to
 * 		transition states.
 *
 * @var system_state_t::collect_housekeeping
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void collect_housekeeping( system_state_t *, command_t * );
 * 			@endcode
 * 		This method will execute the command passed into it ONLY if the state
 * 		allows the collection of all housekeeping data. Otherwise, it does
 * 		nothing.The system_state_t pointer must be the address of the
 * 		structure being used to index the method.
 * 			@code
 * 			var.collect_housekeeping( &var, &command );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * @var system_state_t::transmit_data
 *		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void transmit_data( command_t * );
 * 			@endcode
 * 		This method will execute the command passed into it ONLY if the state
 * 		allows the transmission of data. Otherwise, it does nothing.
 * 		The system_state_t pointer must be the address of the
 * 		structure being used to index the method.
 * 			@code
 * 			var.transmit_data( &var, &command );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * @var system_state_t::handle_MnLP
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void handle_MnLP( command_t * );
 * 			@endcode
 *		This method will execute the command passed into it ONLY if the state
 *		allows IO with the MnLP payload. Otherwise, it does nothing.
 *		The system_state_t pointer must be the address of the
 * 		structure being used to index the method.
 * 			@code
 * 			var.handle_MnLP( &var, &command );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * @var system_state_t::handle_DFGM
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void handle_DFGM( command_t * );
 * 			@endcode
 * 		This method will execute the command passed into it ONLY if the state
 * 		allows IO with the DFGM payload. Otherwise, it does nothing.
 * 		The system_state_t pointer must be the address of the
 * 		structure being used to index the method.
 * 			@code
 * 			var.handle_DFGM( &var, &command );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * @var system_state_t::handle_command_very_high_priority
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void handle_command_very_high_priority( command_t * );
 * 			@endcode
 * 		This method will execute the command passed into it ONLY if the state
 * 		allows the execution of very high priority commands. Otherwise, it
 * 		does nothing. The system_state_t pointer must be the address of the
 * 		structure being used to index the method.
 * 			@code
 * 			var.handle_command_very_high_priority( &var, &command );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * @var system_state_t::handle_command_high_priority
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void handle_command_high_priority( command_t * );
 * 			@endcode
 * 		This method will execute the command passed into it ONLY if the state
 * 		allows the execution of high priority commands. Otherwise, it
 * 		does nothing. The system_state_t pointer must be the address of the
 * 		structure being used to index the method.
 * 			@code
 * 			var.handle_command_high_priority( &var, &command );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * @var system_state_t::handle_command_low_priority
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void handle_command_low_priority( command_t * );
 * 			@endcode
 * 		This method will execute the command passed into it ONLY if the state
 * 		allows the execution of low priority commands. Otherwise, it
 * 		does nothing. The system_state_t pointer must be the address of the
 * 		structure being used to index the method.
 * 			@code
 * 			var.handle_command_low_priority( &var, &command );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 *
 * @var system_state_t::handle_command_very_low_priority
 * 		A method pointer, it is assigned based on the state initialization
 * 		method called:
 * 			@code
 * 			void handle_command_very_low_priority( command_t * );
 * 			@endcode
 * 		This method will execute the command passed into it ONLY if the state
 * 		allows the execution of very low priority commands. Otherwise, it
 * 		does nothing. The system_state_t pointer must be the address of the
 * 		structure being used to index the method.
 * 			@code
 * 			var.handle_command_low_priority( &var, &command );
 * 			@endcode
 * 		In this case, var is being used to index the method, so its address
 * 		is passed to the method.
 */
struct system_state_t
{
	void *				generic_data;
	system_state_id_t 	id;
	entry_handler 		enter_system_state,
						exit_system_state;
	next_state_handler	next_system_state;
	command_handler 	collect_housekeeping,
						transmit_data,
						handle_MnLP,
						handle_DFGM,
						handle_command_very_high_priority,
						handle_command_high_priority,
						handle_command_low_priority,
						handle_command_very_low_priority;
};

/* Used by source file implementations of each state */
#define ENABLE_COMMAND_EXECUTION()	&state_enables_execute
#define DISABLE_COMMAND_EXECUTION() &state_disables_execute
#define NO_ENTRY_EXIT_ROUTINE()		&state_no_entry_no_exit
#define NEVER_EXIT_STATE()			&never_exit_state
void state_enables_execute( system_state_t *, command_t * );
void state_disables_execute( system_state_t *, command_t * );
void state_no_entry_no_exit( system_state_t *, system_state_relay_t * );
system_state_t *never_exit_state(system_state_t *, system_state_relay_t *);

/**
 * @brief
 * 		Gets the id of a state.
 * @details
 * 		Gets the id of a state.	For example,
 * 			@code
 * 			system_state_t state1, state2, state3,
 *
 * 			initialize_power_on_state( &state1 );
 * 			initialize_power_on_state( &state2 );
 * 			initialize_science_state( &state3 );
 *
 * 			printf( "State id: %d, ", query_system_state_id( &state1 ) );
 * 			printf( "State id: %d, ", query_system_state_id( &state2 ) );
 * 			printf( "State id: %d\n", query_system_state_id( &state3 ) );
 * 			@endcode
 * 		will output something similar too:
 * 			@code
 * 				3, 3, 7
 * 			@endcode
 * 		The value of the numbers may be different, but the pattern will
 * 		stay the same.
 * @param state
 * 		The system_state_t object to get the id of.
 * @returns
 * 		The id of the state.
 * @memberof system_state_t
 */
system_state_id_t query_system_state_id( system_state_t *state );

/**
 * @brief
 * 		Initialize a system_state_t structure for use as a power on
 * 		state.
 * @details
 * 		Initialize a system_state_t structure for use as a power on
 * 		state.
 *
 * 		Commands are never executed in this state.
 *
 * 		The next state is the post ejection state only if the post
 * 		ejection pin is released (ie, the satellite is no longer in
 * 		its pod).
 * @param state
 * 		The system_state_t structure to do initialization on.
 * @memberof system_state_t
 */
void initialize_power_on_state( system_state_t *state );

/**
 * @brief
 * 		Initialize a system_state_t structure for use as a post
 * 		ejection state
 *
 * @details
 * 		Initialize a system_state_t structure for use as a post ejection
 * 		state.
 *
 * 		Commands are never executed in this state.
 *
 * 		This state will wait for a duration. The duration waited
 * 		for is defined by the return value of static_timer_t::is_expired
 * 		located in driver_toolkit_t::timer. The driver_toolkit_t structure
 * 		used is the one located in the system_state_relay_t, defined
 * 		at initialization.
 *
 * 		After the duration is up, the next state is start up if there
 * 		is optimal battery power, or pre charge otherwise.
 *
 * @param state
 * 		The system_state_t structure to do initialization on.
 * @memberof system_state_t
 */
void initialize_post_ejection_state( system_state_t *state );

/**
 * @brief
 * 		Initialize a system_state_t structure for use as a pre charge
 * 		state.
 *
 * @details
 * 		Initialize a system_state_t structure for use as a pre charge
 * 		state.
 *
 * 		Commands are never executed in this state.
 *
 * 		The next state is start up only after the battery levels have
 * 		reached optimal level.
 *
 * @param state
 * 		The system_state_t structure to do initialization on.
 *
 * @memberof system_state_t
 */
void initialize_pre_charge_state( system_state_t *state );

/**
 * @brief
 * 		Initialize a system_state_t structure for use as a start up
 * 		state.
 *
 * @details
 * 		Initialize a system_state_t structure for use as a start up
 * 		state.
 *
 * 		Commands are never executed in this state.
 *
 *		This state enables initialization of anything external to the
 *		software (such as subsystems, deployment of payloads, etc).
 *
 * 		The next state is power safe.
 *
 * @param state
 * 		The system_state_t structure to do initialization on.
 *
 * @memberof system_state_t
 */
void initialize_start_up_state( system_state_t *state );

/**
 * @brief
 * 		Initialize a system_state_t structure for use as a power safe
 * 		state.
 *
 * @details
 * 		Initialize a system_state_t structure for use as a power safe.
 * 		state.
 *
 * 		Very high priority commands, housekeeping commands, and
 * 		high priority commands can be executed in this state,
 * 		no other commands can be executed.
 *
 * 		The next state is the detumble state only if the battery level
 * 		is optimal. However, if there is a transmission opportunity and
 * 		the battery level is normal or higher, then the next state is
 * 		transmit state. Transmit state has higher priority than the detumble
 * 		state.
 *
 * @param state
 * 		The system_state_t structure to do initialization on.
 *
 * @memberof system_state_t
 */
void initialize_power_safe_state( system_state_t *state );

/**
 * @brief
 * 		Initialize a system_state_t structure for use as a detumble
 * 		state.
 *
 * @details
 * 		Initialize a system_state_t structure for use as a detumble
 * 		state.
 *
 * 		Very high priority commands, high priority commands, low
 * 		priority commands, and housekeeping commands can be executed
 * 		in this state, no other commands can be executed.
 *
 * 		The next state is the science state only if the adcs has finished
 * 		detumbling and battery levels are optimal. If battery levels fall
 * 		below optimal the next state is powersafe. If the battery levels
 * 		are normal or higher then the next state is transmit. A transistion
 * 		to the transmit state takes priority over all other state
 * 		transistions.
 *
 * @param state
 * 		The system_state_t structure to do initialization on.
 *
 * @memberof system_state_t
 */
void initialize_detumble_state( system_state_t *state );

/**
 * @brief
 * 		Initialize a system_state_t structure for use as a science.
 * 		state.
 *
 * @details
 * 		Initialize a system_state_t structure for use as a science.
 * 		state.
 *
 * 		All commands can be executed in this state.
 *
 * 		The next state, in order of priority is:
 * 			* Transmit state if battery levels are normal or higher
 * 			  and a transmission opportunity is available.
 * 			* Power safe if battery levels fall below optimal.
 * 			* detumble if satellite begins tumbling.
 *
 * @param state
 * 		The system_state_t structure to do initialization on.
 *
 * @memberof system_state_t
 */
void initialize_science_state( system_state_t *state );

/**
 * @brief
 * 		Initialize a system_state_t structure for use as a transmit
 * 		state.
 *
 * @details
 * 		Initialize a system_state_t structure for use as a transmit
 * 		state.
 *
 * 		Very high priority commands, housekeeping commands, and transmit
 * 		commands can be executed in this state, no other commands
 * 		can be executed.
 *
 *		The next state, in order of highest to lowest priority, is:
 *			* power safe if battery level falls below normal and there
 *			  is still a transmission opportunity.
 *			* power safe if battery level is below optimal and there is
 *			  no longer a transmission opportunity.
 *			* detumble if the satellite is tumbling and there is no
 *			  longer a tranmission oppportunity.
 *			* science
 *
 * @param state
 * 		The system_state_t structure to do initialization on.
 *
 * @memberof system_state_t
 */
void initialize_transmit_state( system_state_t *state );

#endif /* INCLUDE_SYSTEM_STATES_H_ */
