/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file commands
 * @author Brendan Bruner
 * @date Jan 28, 2015
 */
#include "commands.h"
#include "system_state_relay.h"
#include "eps.h"
#include "adcs.h"
#include <stdio.h>
#include <stdlib.h>
#include "mission_logger.h"
#include <stdio.h>
#include <string.h>

/************************************************************************/
/* Private, non virtual, method implementations							*/
/************************************************************************/
#define generic_init( command )	command->__execution_status = PENDING

/************************************************************************/
/* Public, non virtual, method implementations							*/
/************************************************************************/
void execute_command( command_t *command )
{
	/* TODO: SD card logging of command execution. */
	/* Do some logging to SD card. */
	if( command != NULL && command->execute != NULL )
	{
		command->__execution_status = EXECUTING;
		command->execute( command );
		command->__execution_status = EXECUTED;
	}
	/* Do some more logging to SD card. */
}

command_status_t command_status( command_t *command )
{
	return command->__execution_status;
}


/************************************************************************/
/* Set state commands													*/
/************************************************************************/
static void execute_set_state_power_safe( command_t *self )
{
	system_state_relay_t *relay;

	relay = (system_state_relay_t *) self->generic_data;
	relay_force_change_system_state( relay, &relay->power_safe );
}
void initialize_command_set_state_power_safe( 	command_t *self,
												system_state_relay_t *data )
{
	generic_init( self );
	self->generic_data = (void *) data;
	self->execute = &execute_set_state_power_safe;
}


/************************************************************************/
/* Housekeeping commands												*/
/************************************************************************/
/* ----------------------------EPS------------------------------------- */
static void execute_collect_eps_hk( command_t *self )
{
	eps_t *eps;
	char const *str;
	mission_logger_t *logger;

	eps = &((driver_toolkit_t *) self->generic_data)->eps;
	logger = &((driver_toolkit_t *) self->generic_data)->logger;

	eps->hk_refresh(eps);
	str = eps_mode_to_name[eps->mode( eps )];
	write_eps_hk( logger, str, strlen( str ) );
}
void initialize_command_collect_eps_hk( command_t *self, driver_toolkit_t *kit )
{
	generic_init( self );
	self->generic_data = (void *) kit;
	self->execute = &execute_collect_eps_hk;
}

/* ----------------------------ADCS------------------------------------- */
static void execute_collect_adcs_hk( command_t *self)
{
	adcs_t *adcs;
	mission_logger_t *logger;
	char const *str;

	adcs = &((driver_toolkit_t *) self->generic_data)->adcs;
	logger = &((driver_toolkit_t *) self->generic_data)->logger;
	str = adcs_mode_to_name[adcs->mode( adcs )];

	write_adcs_hk( logger, str, strlen( str ) );
}
void initialize_command_collect_adcs_hk( command_t *command, driver_toolkit_t *kit )
{
	generic_init( command );
	command->generic_data = (void *) kit;
	command->execute = &execute_collect_adcs_hk;
}

/* ----------------------------COMM------------------------------------- */
static void execute_collect_comm_hk( command_t *self )
{
	comm_t *comm;
	mission_logger_t *logger;
	nanocom_hk_t comm_hk[HK_ARRAY_SIZE];
	uint8_t num_received;

	comm = &((driver_toolkit_t *) self->generic_data)->comm;
	logger = &((driver_toolkit_t *) self->generic_data)->logger;
	comm->com_get_hk( comm, comm_hk, &num_received );

	for( ; num_received > 0; --num_received )
	{
		write_comm_hk( logger, (char const *) &comm_hk[num_received-1], sizeof( nanocom_hk_t ) );
	}
}
void initialize_command_collect_comm_hk( command_t *command, driver_toolkit_t *kit )
{
	generic_init( command );
	command->generic_data = (void *) kit;
	command->execute = &execute_collect_comm_hk;
}

/************************************************************************/
/* Transmit Commands													*/
/************************************************************************/
#define READ_LEN 40
/* ----------------------------EPS------------------------------------- */
static void transmit_eps_hk_execute( command_t *command )
{
	mission_logger_t *logger;
	comm_t *comm;
	char eps_buff[READ_LEN];
	unsigned int read;

	logger = &((driver_toolkit_t *) command->generic_data)->logger;
	comm = &((driver_toolkit_t *) command->generic_data)->comm;

	read = next_eps_hk( logger, eps_buff+1, READ_LEN-1 );
	if( read > 0 )
	{
		eps_buff[0] = 0;
		comm->transmit( comm, (uint8_t *) eps_buff, READ_LEN );
	}
}
void initialize_command_transmit_eps_hk(	command_t *command,
											driver_toolkit_t *kit )
{
	generic_init( command );
	command->generic_data = (void *) kit;
	command->execute = &transmit_eps_hk_execute;
}

/* ----------------------------ADCS------------------------------------ */
static void transmit_adcs_hk_execute( command_t *command )
{
	char adcs_buff[READ_LEN];
	mission_logger_t *logger;
	unsigned int read;
	comm_t *comm;

	logger = &((driver_toolkit_t *) command->generic_data)->logger;
	comm = &((driver_toolkit_t *) command->generic_data)->comm;

	read = next_adcs_hk( logger, adcs_buff+1, READ_LEN-1 );
	if( read > 0 )
	{
		adcs_buff[0] = 1;
		comm->transmit( comm, (uint8_t *) adcs_buff, READ_LEN );
	}

}
void initialize_command_transmit_adcs_hk(	command_t *command,
											driver_toolkit_t *kit )
{
	generic_init( command );
	command->generic_data = (void *) kit;
	command->execute = &transmit_adcs_hk_execute;
}

/* ----------------------------COMM------------------------------------ */
static void transmit_comm_hk_execute( command_t *command )
{
	char comm_buff[READ_LEN];
	mission_logger_t *logger;
	comm_t *comm;
	unsigned int read;

	logger = &((driver_toolkit_t *) command->generic_data)->logger;
	comm = &((driver_toolkit_t *) command->generic_data)->comm;

	read = next_comm_hk( logger, comm_buff+1, READ_LEN-1 );
	if( read > 0 )
	{
		comm_buff[0] = 2;
		comm->transmit( comm, (uint8_t *) comm_buff, READ_LEN );
	}
}
void initialize_command_transmit_comm_hk(	command_t *command,
											driver_toolkit_t *kit )
{
	generic_init( command );
	command->generic_data = (void *) kit;
	command->execute = &transmit_comm_hk_execute;
}

/* ----------------------------STATE---------------------------------- */
static void transmit_state_hk_execute( command_t *command )
{
	char state_buff[READ_LEN];
	mission_logger_t *logger;
	comm_t *comm;
	unsigned int read;

	logger = &((driver_toolkit_t *) command->generic_data)->logger;
	comm = &((driver_toolkit_t *) command->generic_data)->comm;

	read = next_state_hk( logger, state_buff+1, READ_LEN-1 );
	if( read > 0 )
	{
		state_buff[0] = 3;
		comm->transmit( comm, (uint8_t *) state_buff, READ_LEN );
	}
}
void initialize_command_transmit_state_hk(	command_t *command,
											driver_toolkit_t *kit )
{
	generic_init( command );
	command->generic_data = (void *) kit;
	command->execute = &transmit_state_hk_execute;
}

/************************************************************************/
/* Commands for Testing Only											*/
/************************************************************************/
static void execute_print_string( command_t *command )
{
	printf( (char *) command->generic_data );
}
void initialize_command_print_string( command_t *command, char const *data )
{
	generic_init( command );
	command->generic_data = (void *) data;
	command->execute = &execute_print_string;
}

static void send_time_to_adcs_over_i2c(command_t *self)
{
	adcs_t *adcs;
	adcs_unix_time_t time;
	time.unix_time = 987654321;
	time.millis = 54321;
	adcs = &((driver_toolkit_t *) self->generic_data)->adcs;
	adcs->adcs_set_unix_time(adcs, &time);
}

static void receive_time_from_adcs_over_i2c(command_t *self)
{
	adcs_t *adcs;
	adcs = &((driver_toolkit_t *) self->generic_data)->adcs;
	adcs_unix_time_t *time = adcs->adcs_send_curr_time(adcs);
	printf("unix_time: %d\n", time->unix_time);
	printf("millis: %d\n", time->millis);
	printf("\n");
	free(time);
}

void initialize_command_send_time_to_adcs_over_i2c( command_t *self, driver_toolkit_t *drivers )
{
	self->generic_data = (void *) drivers;
	self->execute = &send_time_to_adcs_over_i2c;
}

void initialize_command_receive_time_from_adcs_over_i2c( command_t *self, driver_toolkit_t *drivers )
{
	self->generic_data = (void *) drivers;
	self->execute = &receive_time_from_adcs_over_i2c;
}

static void execute_print_state( command_t *command )
{
	system_state_relay_t *relay;

	relay = (system_state_relay_t *) command->generic_data;
	printf( "\nCurrent state: %s\n", state_id_to_name[relay_system_state_id( relay )] );
}
void initialize_command_print_state( command_t *command, system_state_relay_t *relay )
{
	generic_init( command );
	command->generic_data = (void *) relay;
	command->execute = &execute_print_state;
}
