/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file system_state_relay
 * @author Brendan Bruner
 * @date Jan 29, 2015
 */

#include "core_defines.h"
#include "system_state_relay.h"
#include "system_states.h"
#include "mission_logger.h"
#include "commands.h"
#include <stdio.h>

#ifndef NULL
#define NULL (void *) 0
#endif

/************************************************************************/
/* Private Defines														*/
/************************************************************************/
#define generic_relay( relay_p, command_p, handle_name )				\
				do {													\
					system_state_t *state;								\
																		\
					relay_next_state( relay_p );						\
																		\
					xSemaphoreTake( relay_p->__state_mutex, portMAX_DELAY ); \
					state = relay_p->current_state;						\
					xSemaphoreGive( relay_p->__state_mutex );			\
																		\
					if( command_p != NULL )								\
					{													\
						state-> handle_name ( state, command_p );		\
					}													\
				} while( 0 )


/************************************************************************/
/* State Manipulation Methods											*/
/************************************************************************/
/* Private function to change state. */
static void relay_next_state( system_state_relay_t *relay )
{
	system_state_t *next_state, *current_state;

	xSemaphoreTake( relay->__state_mutex, portMAX_DELAY );
	current_state = relay->current_state;
	xSemaphoreGive( relay->__state_mutex );

	next_state = current_state->next_system_state( current_state, relay );

	if ( next_state == NULL )
	{
		return;
	}
	else
	{
		xSemaphoreTake( relay->__state_mutex, portMAX_DELAY );
		relay->current_state = next_state;
		xSemaphoreGive( relay->__state_mutex );

		current_state->exit_system_state( current_state, relay );
		next_state->enter_system_state( next_state, relay );

		log_state_change( &relay->drivers->logger, query_system_state_id( next_state ) );
#ifdef RELAY_DEMO_CODE
		printf( "\nCurrent state: %s\n", state_id_to_name[relay_system_state_id( relay )] );
#endif
	}
}

void relay_force_change_system_state( 	system_state_relay_t *relay,
										system_state_t *next )
{
	if( next == NULL )
	{
		return;
	}
	relay->current_state->exit_system_state( relay->current_state, relay);

	xSemaphoreTake( relay->__state_mutex, portMAX_DELAY );
	relay->current_state = next;
	xSemaphoreGive( relay->__state_mutex );

	relay->current_state->enter_system_state( relay->current_state, relay );
}

/************************************************************************/
/* Methods to relay commands from application to state.					*/
/************************************************************************/

void relay_command_very_high_priority( 	system_state_relay_t *relay,
										command_t *command )
{
	/* Make sure the current state of the system is most up to date before
	 * attempting to relay the command. */
	relay_next_state( relay );
	if( command != NULL )
	{
		relay->current_state->handle_command_very_high_priority( relay->current_state, command );
	}
}

void relay_command_high_priority(	system_state_relay_t *relay,
									command_t *command )
{
	/* Make sure the current state of the system is most up to date before
	 * attempting to relay the command. */
	relay_next_state( relay );
	if( command != NULL )
	{
		relay->current_state->handle_command_high_priority( relay->current_state, command );
	}
}

void relay_command_low_priority(	system_state_relay_t *relay,
									command_t *command )
{
	/* Make sure the current state of the system is most up to date before
	 * attempting to relay the command. */
	relay_next_state( relay );
	if( command != NULL )
	{
		relay->current_state->handle_command_low_priority( relay->current_state, command );
	}
}

void relay_command_very_low_priority( 	system_state_relay_t *relay,
										command_t *command )
{
	/* Make sure the current state of the system is most up to date before
	 * attempting to relay the command. */
	relay_next_state( relay );
	if( command != NULL )
	{
		relay->current_state->handle_command_very_low_priority( relay->current_state, command );
	}
}

void relay_command_collect_housekeeping(	system_state_relay_t *relay,
											command_t *command )
{
	/* Make sure the current state of the system is most up to date before
	 * attempting to relay the command. */
	/*relay_next_state( relay );
	if( command != NULL )
	{
		relay->current_state->collect_housekeeping( relay->current_state, command );
	}*/
	generic_relay( relay, command, collect_housekeeping );
}

void relay_command_transmit_data(	system_state_relay_t *relay,
									command_t *command )
{
	/* Make sure the current state of the system is most up to date before
	 * attempting to relay the command. */
	relay_next_state( relay );
	if( command != NULL )
	{
		relay->current_state->transmit_data( relay->current_state, command );
	}
}

void relay_command_handle_MnLP(	system_state_relay_t *relay,
								command_t *command )
{
	/* Make sure the current state of the system is most up to date before
	 * attempting to relay the command. */
	relay_next_state( relay );
	if( command != NULL )
	{
		relay->current_state->handle_MnLP( relay->current_state, command );
	}
}

void relay_command_handle_DFGM(	system_state_relay_t *relay,
								command_t *command )
{
	/* Make sure the current state of the system is most up to date before
	 * attempting to relay the command. */
	relay_next_state( relay );
	if( command != NULL )
	{
		relay->current_state->handle_DFGM( relay->current_state, command );
	}
}



/************************************************************************/
/* Methods to relay data from state back to application.				*/
/************************************************************************/
system_state_id_t relay_system_state_id( system_state_relay_t *relay )
{
	return relay->current_state->id;
}

/************************************************************************/
/* Initialization Methods												*/
/************************************************************************/
void initialize_system_state_relay( system_state_relay_t 	*relay,
									system_state_t 			*initial_state,
									driver_toolkit_t 		*drivers )
{
	initialize_power_on_state( &relay->power_on );
	initialize_post_ejection_state( &relay->post_ejection );
	initialize_pre_charge_state( &relay->pre_charge );
	initialize_start_up_state( &relay->start_up );
	initialize_power_safe_state( &relay->power_safe );
	initialize_detumble_state( &relay->detumble );
	initialize_science_state( &relay->science );
	initialize_transmit_state( &relay->transmit );

	relay->drivers = drivers;

	relay->__state_mutex = xSemaphoreCreateMutex();
	if( relay->__state_mutex == NULL )
	{
		/* TODO: error handle failed mutex creation. */
	}
	xSemaphoreGive( relay->__state_mutex );

	if( initial_state == NULL )
	{
		/* No initial state specified, default to power on. */
		relay->current_state = &relay->power_on;
	}
	else
	{
		relay->current_state = initial_state;
	}
	relay->current_state->enter_system_state( relay->current_state, relay );
}
