/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file comm
 * @author Brendan Bruner
 * @author Jeff Ryan
 * @date Feb 25, 2015
 */

#ifndef INCLUDE_COMM_H_
#define INCLUDE_COMM_H_

#include <stdint.h>
#include "mock_comms_structs.h"
#include "telecommand_packet.h"

/* Required typedefs */
typedef struct comm_t comm_t;
typedef unsigned char comm_beacon_t;

#define COMM_IS_BEACONED	((comm_beacon_t) 0)
#define COMM_NOT_BEACONED	((comm_beacon_t) 1)
extern char const * const comm_mode_to_name[]; /* declare */

#define NO_TELECOMMAND			0
#define INCOMING_TELECOMMAND	1

#define HK_ARRAY_SIZE		8

#define BEACON_REQUEST	0
#define COMM_HK_REQUEST	1
#define COMM_TC_REQUEST	2

/**
 * @struct comm_t
 * @brief
 * 		Structure used with API for communications board.
 *
 * @details
 * 		Structure used with API for communications board. All instances
 * 		of this structure must be initialized before use. For example,
 * 			@code
 * 			initialize_comm_nanocomm( comm );
 * 			@endcode
 * 		In this code snippet, the variable "comm" is of type comm_t. The
 * 		function called will initialize the structure to do IO with
 * 		gomespace's nanocomm board.
 *
 * 		There are several different initialization functions depending on
 * 		the needs of the application code.
 *
 * @attention
 * 		Attention to code developers!
 * 		<br>Multiple instance of this structure - initialized the same way, MUST
 * 		function correctly. In other words, there cannot be a restriction on
 * 		the number of structures that exist.
 *
 * @var comm_t::beacon
 * 		A function pointer with the prototype:
 * 			@code
 * 			comm_beacon_t beacon( comm_t * );
 * 			@endcode
 * 		The method's input is the comm_t structure used to access it. The
 * 		return value is one of:
 * 			@code
 * 			COMM_IS_BEACONED
 * 			COMM_NOT_BEACONED
 * 			@endcode
 * 		When COMM_IS_BEACONED is returned this means that some third party
 * 		is beaconing the communications hardware or has very recently
 * 		beaconed the communications hardware. If COMM_NOT_BEACONED is
 * 		returned, this means the communications hardware is currently not
 * 		being beaconed and hasn't been beaconed recently.
 *
 * @var comm_t::com_set_conf
 * 		A function pointer with prototype:
 * 			@code
 * 			int com_set_conf( comm_t *, nanocom_conf_t * );
 * 			@endcode
 * 		The first argument is a pointer to the comm_t structure used to access this pointer.
 * 		<br>The second argument is a pointer to the configuration file to be used.
 * 		<br>Returns the CSP return value.
 * 		<br>This takes the pointer (nanocom_conf_t *) to the config file and sets the comm board's
 * 		configuration file with that data.
 *
 *
 *
 * @var comm_t::com_get_conf
 * 		A function pointer with prototype:
 * 			@code
 * 			int com_get_conf( comm_t *, nanocom_conf_t * );
 * 			@endcode
 * 		The first argument is a pointer to the comm_t structure used to access this pointer.
 * 		<br>The second argument is a pointer to the location to store the configuration file.
 * 		<br>Returns the CSP return value.
 * 		<br>This saves the comm board's current configuration file to the pointer (nanocom_conf_t *)
 * 		given.
 *
 * 	@var comm_t::com_restore_conf
 * 		A function pointer with prototype:
 * 			@code
 * 			int com_restore_conf( comm_t * );
 * 			@endcode
 * 		The first argument is a pointer to the comm_t structure used to access this pointer.
 * 		<br>Returns the CSP return value.
 * 		<br>It deletes the comm board's configuration file and restores it to default settings.
 *
 * 	@var comm_t::com_get_status
 * 		A function pointer with prototype:
 * 			@code
 * 			int com_get_status( comm_t *, nanocom_data_t * );
 * 			@endcode
 * 		The first argument is a pointer to the comm_t structure used to access this pointer.
 * 		<br>The second argument is a pointer to the location to store the status data.
 * 		<br>Returns the CSP return value.
 * 		<br>It gets and saves the status of the comms board to the pointer (nanocom_data_t *) given.
 *
 * 	@var comm_t::com_get_log_rssi
 * 		A function pointer with prototype:
 * 			@code
 * 			int com_get_log_rssi( comm_t *, nanocom_rssi_t *, uint8_t * );
 * 			@endcode
 * 		The first argument is a pointer to the comm_t structure used to access this pointer.
 * 		<br>The second argument is a pointer to where the array (of nanocom_rssi_t) will be stored (size >=10).
 * 		<br>The third argument is a pointer to the number of rssi structs received.
 * 		<br>Returns the CSP return value.
 * 		<br>It gets and saves arrays of the rssi info to the pointer (nancom_rssi_t *) given.
 * 		It also saves the number of rssi structs at the pointer (uint8_t *) given.
 *
 * 	@var comm_t::com_get_hk
 * 		A function pointer with prototype:
 * 			@code
 * 			int com_get_hk( comm_t *, nanocom_hk_t *, uint8_t * );
 * 			@endcode
 * 		The first argument is a pointer to the comm_t structure used to access this pointer.
 * 		<br>The second argument is a pointer to where the array will be stored (size =8 ).
 * 		<br>The third argument is a pointer to the number of hk entries received.
 * 		<br>Returns the CSP return value.
 * 		<br>It gets and saves an array of the housekeeping info to the pointer (nanocom_hk_t *) given.
 * 		It also saves the number of hk entries received at the pointer (uint8_t *) given.
 *
 * @var comm_t::poll_telecommand
 * 		A function pointer with prototype
 * 			@code
 * 			uint8_t poll_telecommand( comm_t *comm, telecommand_packet_t *command );
 * 			@endcode
 * 		This method polls for any incomming telecommands from ground stations.
 *
 * 		<b>Parameters</b>
 * 		<br>comm: A pointer to the comm_t structure used to access this method.
 * 		<br>command: A pointer to the telecommand packet.
 *
 * 		<b>Returns</b>
 * 		<br>NO_TELECOMMAND is returned if there is no telecommand. The command
 * 		pointer is void.
 * 		<br>INCOMING_TELECOMMAND if a telecommand has been buffered.
 *
 * @var comm_t::transmit
 * 		A function pointer with prototype:
 * 			@code
 * 			uint8_t transmit( comm_t *comm, uint8_t *data,  uint16_t size );
 * 			@endcode
 * 		The method transmits data to ground station (through comms board).
 *
 * 		<b>Parameters</b>
 * 		<br>comm: A pointer to the comm_t structure used to access the method.
 * 		<br>data: A pointer to the byte array to transmit.
 * 		<br>size: The number of bytes to transmit.
 *
 * 		<b>Returns</b>
 * 		<br>The number of bytes transmitted. If zero bytes were transmitted, there
 * 		is no ground station in range.
 */
struct comm_t
{

	nanocom_conf_t nanocom_conf_struct;

	void *data;
	comm_beacon_t (*beacon)( comm_t * );
	int (*com_set_conf)( comm_t *, nanocom_conf_t * );
	int (*com_get_conf)( comm_t *, nanocom_conf_t * );
	int (*com_restore_conf)( comm_t * );
	int (*com_get_status)(comm_t *, nanocom_data_t * );
	int (*com_get_log_rssi)(comm_t *, nanocom_rssi_t *, uint8_t * );
	int (*com_get_hk)(comm_t *, nanocom_hk_t *, uint8_t * );
	uint8_t (*poll_telecommand)( comm_t *, telecommand_packet_t * );
	uint8_t (*transmit)( comm_t *, uint8_t *, uint16_t );
};

/**
 * @brief
 * 		Initialize a software mock up of the communications hardware.
 *
 * @details
 * 		Initialize a software mock up of the communications hardware.
 * 		Multiple instances of comm_t structures can be safetly intialized
 * 		with this method.
 *
 * 		The method,
 * 			@code
 * 			comm_beacon_t comm_t::beacon( comm_t * );
 * 			@endcode
 * 		will initially return COMM_NOT_BEACONED for a long duration, then
 * 		return COMM_IS_BEACONED for a short duration. Rince and repeat.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_set_conf( comm_t *, nanocom_conf_t * );
 * 			@endcode
 * 		will set a config file determined by the nancom_conf_t pointer.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_get_conf( comm_t *, nanocom_conf_t * );
 * 			@endcode
 * 		will get the current config file.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_restore_conf( comm_t * );
 * 			@endcode
 * 		Will restore config files to factory (made up numbers, mostly 0s).
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_get_status( comm_t *, nanocom_data_t * );
 * 		Will get the board status.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_get_log_rssi( comm_t *, nanocom_rssi_t *, uint8_t * );
 * 		Will get the rssi logs.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_get_hk( comm_t *, nanocom_hk_t *, uint8_t * );
 * 		Will get the hk entries.
 *
 * @param comm
 * 		A pointer to the comm_t structure to initialize.
 *
 * @memberof comm_t
 */
void initialize_comm_lpc_local( comm_t *comm );

/**
 * @brief
 * 		Initialize a software mock up of the communications hardware on
 * 		an arduino.
 *
 * @details
 * 		Initialize a software mock up of the communications hardware on an
 * 		arduino. Communication between lpc and arduino is over uart.
 * 		Multiple instances of comm_t structures can be safetly intialized
 * 		with this method.
 *
 * 		The method,
 * 			@code
 * 			comm_beacon_t comm_t::beacon( comm_t * );
 * 			@endcode
 * 		will initially return COMM_NOT_BEACONED for a long duration, then
 * 		return COMM_IS_BEACONED for a short duration. Rince and repeat.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_set_conf( comm_t *, nanocom_conf_t * );
 * 			@endcode
 * 		will set a config file determined by the nancom_conf_t pointer.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_get_conf( comm_t *, nanocom_conf_t * );
 * 			@endcode
 * 		will get the current config file.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_restore_conf( comm_t * );
 * 			@endcode
 * 		Will restore config files to factory (made up numbers, mostly 0s).
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_get_status( comm_t *, nanocom_data_t * );
 * 		Will get the board status.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_get_log_rssi( comm_t *, nanocom_rssi_t *, uint8_t * );
 * 		Will get the rssi logs.
 *
 * 		The method,
 * 			@code
 * 			int comm_t::com_get_hk( comm_t *, nanocom_hk_t *, uint8_t * );
 * 		Will get the hk entries.
 *
 * @param comm
 * 		A pointer to the comm_t structure to initialize.
 *
 * @memberof comm_t
 */
void initialize_comm_arduino_remote( comm_t *comm );

#endif /* INCLUDE_COMM_H_ */
