/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file driver_toolkit
 * @author Brendan Bruner
 * @date Feb 4, 2015
 */
#ifndef INCLUDE_DRIVER_TOOLKIT_H_
#define INCLUDE_DRIVER_TOOLKIT_H_

#include "eps.h"
#include "ejection_pin.h"
#include "static_timer.h"
#include "comm.h"
#include "adcs.h"
#include "mission_logger.h"

/* typedefs */
typedef struct driver_toolkit_t driver_toolkit_t;

/**
 * @struct driver_toolkit_t
 * @brief
 * 		A container for all drivers used by commands, states, etc.
 *
 * @details
 * 		A container for all drivers used by commands, states, etc. The
 * 		driver_toolkit_t structure must be initialized. There are several
 * 		initialization functions. Most of them have the generic form as
 * 		below,
 * 			@code
 * 			initialize_driver_toolkit( driver_toolkit_t * );
 * 			@endcode
 * 		This method sets up the toolkit with all the drivers.
 *
 * @var driver_toolkit_t::eps
 * 		An eps_t driver. This variable can be read, but should
 * 		never be written to.
 *
 * @var driver_toolkit_t::ejection_pin
 * 		A driver for the post ejection pin. This variable can be read,
 * 		but should not be written to.
 *
 * @var driver_toolkit_t::timer
 * 		Driver used for timing a countdown.
 *
 * @var driver_toolkit_t::logger
 * 		Driver for logging data.
 */
struct driver_toolkit_t
{
	eps_t eps;
	ejection_pin_t ejection_pin;
	static_timer_t timer;
	comm_t comm;
	adcs_t adcs;
	mission_logger_t logger;
};

/**
 * @brief
 * 		Initializes the driver_toolkit_t structure to use mock drivers
 * 		on the lpc board.
 *
 * @details
 * 		Initializes the driver_toolkit_t structure to use mock drivers
 * 		on the lpc board.
 * 			The eps_t, comm_t, and adcs_t drivers are initialized as
 * 			an lpc_local driver.
 * 			<br>The ejction_pin is initialized as a mock up.
 * 			<br>The timer is initialized as a mock up.
 * 			<br>The logger is initialized to do IO with an SD card
 *
 * @param toolkit
 * 		A pointer to the driver_toolkit_t structure to be initialized.
 *
 * @memberof driver_toolkit_t
 */
void initialize_driver_toolkit_lpc_local( driver_toolkit_t *toolkit );

#endif /* INCLUDE_DRIVER_TOOLKIT_H_ */
