/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file sd
 * @author Brendan Bruner
 * @date Mar 12, 2015
 */
#ifndef INCLUDE_MISSION_LOGGER_H_
#define INCLUDE_MISSION_LOGGER_H_

#include "../FatFS/src/ff.h"
#include <stdint.h>

#define BEGINNING_OF_FILE	0

/* Required typedefs */
typedef struct mission_logger_t mission_logger_t;
typedef enum
{
	MISSION_LOGGER_OK,
	MISSION_LOGGER_FAILURE,
	MISSION_LOGGER_OVERWRITE_FAILURE,
	MISSION_LOGGER_NO_FILE,
	MISSION_LOGGER_END_OF_FILE
} ml_result_t;

/**
 * @struct mission_logger_t
 *
 * @brief
 * 		A structure used to access file IO API.
 *
 * @details
 * 		A structure used to access file IO API. This structure maintains state
 * 		data and puts mutual exclusion on files.
 *
 * @remark
 * 		Any method which returns,
 * 			MISSION_LOGGER_FAILURE
 * 		will reject changes made to the file. There will not be data corruption
 * 		when this error code is returned.
 *
 * @var mission_logger_t::open_a
 * 		A function pointer with prototype:
 * 			@code
 * 			ml_result_t open_a( mission_logger_t *sd, const char *name );
 * 			@endcode
 * 		This method opens a file which is named "name" for writing. The position
 * 		of the cursor (ie, where writing will start) is placed a the end of
 * 		the file. If a file is already opened the method will block until the
 * 		currently open file is closed.
 *
 * 		Files opened this way can also be read from.
 *
 * 		<b>Parameters</b>
 *			<br>sd: A pointer to the mission_logger_t structure used to access method.
 * 			<br>name:	Name of the file to open. For example, "hk_eps.txt"
 *
 * 		<b>Returns</b>
 * 			<br>MISSION_LOGGER_OK			if the file opened successfully
 * 			<br>MISSION_LOGGER_FAILURE		if the file failed to open
 *
 * @var mission_logger_t::open_r
 * 		A function pointer with prototype:
 * 			@code
 * 			ml_result_t open_r( mission_logger_t *sd, const char *name );
 * 			@endcode
 * 		This method opens the file "name" for reading. The cursor (ie, the place
 * 		where data is read from) is placed at the very beginning of the file. If
 * 		a file is already open, this method blocks until the currently open file
 * 		is closed.
 *
 * 		Files opened this way cannot be written to.
 *
 * 		<b>Parameters</b>
 * 			<br>sd: A pointer to the mission_logger_t structure used to access the method.
 * 			<br>name: The name of the file to open for reading
 *
 * 		<b>Returns</b>
 * 			<br>MISSION_LOGGER_OK			if the file opened successfully
 * 			<br>MISSION_LOGGER_FAILURE		if the file failed to open
 *
 * @var mission_logger_t::close
 * 		A function pointer with prototype:
 * 			@code
 * 			ml_result_t close( mission_logger_t *sd );
 * 			@endcode
 * 		this method closes the currently open file. If no file is opened it
 * 		does nothing.
 *
 *		<b>Paramters</b>
 *			<br>sd: A pointer to the mission_logger_t structure used to access method.
 *
 * 		<b>Returns</b>
 * 			<br>MISSION_LOGGER_OK		if the file closed successfully
 * 			<br>MISSION_LOGGER_NO_FILE	if there was no file to close
 * 			<br>MISSION_LOGGER_FAILURE	if the file failed to close
 *
 * @var mission_logger_t::file_size
 * 		A function pointer with prototype,
 * 			@code
 * 			ml_result_t file_size( mission_logger_t *sd, unsigned long *size );
 * 			@endcode
 * 		This method queries the size of the currently open file.
 *
 * 		<b>Parameters</b>
 * 			<br>sd: A pointer to the mission_logger_t structure used to access this method
 * 			<br>size: A pointer to the variable to save the size of the file to. The size
 * 			is in bytes.
 *
 * 		<b>Returns</b>
 * 			<br>MISSION_LOGGER_OK		if successful
 * 			<br>MISSION_LOGGER_FAILURE	if the method failed
 * 			<br>MISSION_LOGGER_NO_FILE	if there was no file open to query
 *
 * @var mission_logger_t::write
 * 		A function pointer with prototype,
 * 			@code
 * 			ml_result_t write( mission_logger_t *sd, const char *msg, unsigned int size, unsigned int *written);
 * 			@endcode
 * 		This method will write a string to the currently open file. The write
 * 		will begin where the cursor is in the file. After the write, the cursor
 * 		will be moved to the end of the write. For example,
 *
 * 			example_file.txt
 * 			this is data |
 *
 * 		Say that the "|" is the cursor and we write the string "derp cat". The
 * 		file will look like this,
 *
 * 			example_file.txt
 * 			this is data derp cat|
 *
 *		If the currently open file is being written to in a different process /
 *		task the method blocks until that write finishes.
 *
 *		Data is interpretted as ASCII and written as its ASCII equivalent. For
 *		example, writing a "\n" will write a new line, not a '\' and an 'n'.
 *
 *		<b>Parameters</b>
 *			<br>sd: A pointer to the mission_logger_t structure used to access
 *			this method.
 *			<br>msg: The string to write to the file. Must be null terminated.
 *			Writing will stop when the null terminating character is reached.
 *			<br>size: The maximum number of bytes to write to the file. Writing
 *			will stop when this size is exceeded. This has a higher priority than
 *			the null terminating character.
 *			<br>written: Pointer to location where number of bytes actually written
 *			is saved.
 *
 *		<b>Returns</b>
 *			<br>MISSION_LOGGER_OK		if the write was successful
 *			<br>MISSION_LOGGER_FAILURE	if the write failed. The open file
 *			will be closed and must be reopened to do more writes.
 *			<br>MISSION_LOGGER_NO_FILE	if there was no file open to write to
 *
 * @var mission_logger_t::write_stream
 * 		A function pointer with prototype,
 * 			@code
 * 			ml_result_t write_stream( mission_logger_t *sd, const uint8_t *msg, unsigned int size, unsigned int *written);
 * 			@endcode
 * 		This method will write a string to the currently open file. The write
 * 		will begin where the cursor is in the file. After the write, the cursor
 * 		will be moved to the end of the write. For example,
 *
 * 			example_file.txt
 * 			this is data |
 *
 * 		Say that the "|" is the cursor and we write the string "derp cat". The
 * 		file will look like this,
 *
 * 			example_file.txt
 * 			this is data derp cat|
 *
 *		If the currently open file is being written to in a different process /
 *		task the method blocks until that write finishes.
 *
 *		Data is interpreted as a byte stream. For example, writing a "\n" will
 *		be interpreted as a ten. '10' will be written.
 *
 *		<b>Parameters</b>
 *			<br>sd: A pointer to the mission_logger_t structure used to access
 *			this method.
 *			<br>msg: The string to write to the file. Must be null terminated.
 *			Writing will stop when the null terminating character is reached.
 *			<br>size: The maximum number of bytes to write to the file. Writing
 *			will stop when this size is exceeded. This has a higher priority than
 *			the null terminating character.
 *			<br>written: Pointer to location where number of bytes actually written
 *			is saved.
 *
 *		<b>Returns</b>
 *			<br>MISSION_LOGGER_OK		if the write was successful
 *			<br>MISSION_LOGGER_FAILURE	if the write failed. The open file
 *			will be closed and must be reopened to do more writes.
 *			<br>MISSION_LOGGER_NO_FILE	if there was no file open to write to
 *
 *
 * @var mission_logger_t::read
 * 		A function pointer with prototype
 * 			@code
 * 			ml_result_t read( mission_logger_t *sd, const char *msg, unsigned int size, unsigned int *read);
 * 			@endcode
 * 		Read from the currently open file. The read will start where the cursor
 * 		is in the file and continue for "size" bytes, the cursor will move with the read.
 * 		For example, if "size" is two, it will read the first two bytes after (to the
 * 		of) the cursor. The cursor will then be moved two bytes over (to the right). If the
 * 		end of file is reached before "size" bytes are read, the read will stop.
 *
 *		<b>Parameters</b>
 *			<br>sd: A pointer to the structure used to access this method
 *			<br>msg: A pointer to an array to put the read bytes. Will always
 *			be given a null terminating character
 *			<br>size: The number of bytes to read. Must be less than or equal
 *			to the size of the array msg points to.
 *			<br>read: The number of bytes actually read.
 *
 *		<b>Returns</b>
 *			<br>MISSION_LOGGER_OK		if successful
 *			<br>MISSION_LOGGER_FAILURE	if the method failed. The array pointed to by
 *			msg may contain garbage. The currently open file will be closed and must
 *			be reopened.
 *			<br>MISSION_LOGGER_NO_FILE	No file was open to read from. Nothing was
 *			done.
 *			<br>MISSION_LOGGER_END_OF_FILE	end of file reached, number bytes read
 *			may be different the number of bytes requested to be read.
 *
 * @var mission_logger_t::go_to
 * 		A pointer to a method with prototype,
 * 			@code
 * 			ml_result_t go_to( mission_logger_t *sd, unsigned int position );
 * 			@endcode
 * 		Will put the cursor at the absolute position specified by the input
 * 		argument "position". "position" is the number of byte from the beginning
 * 		of the file to place the cursor. For example, if this method is called
 * 		with "position" being four,
 *
 * 			The |cursor is back there.
 *
 * 		Where "|" is the cursor.
 *
 *		<b>Parameters</b>
 *			<br>sd: A pointer to the structure used to access this method
 *			<br>position: The number of bytes from the beginning of the file
 *			to put the cursor
 *
 *		<b>Returns</b>
 *			<br>MISSION_LOGGER_OK		if successful
 *			<br>MISSION_LOGGER_FAILURE	if the method failed. The currently
 *			open file will be closed and must be reopened.
 *			<br>MISSION_LOGGER_NO_FILE	No file was open. Nothing was done.
 *
 * @var mission_logger_t::delete
 * 		A function pointer with prototype,
 * 			@code
 * 			ml_result_t delete( mission_logger_t *sd );
 * 			@endcode
 * 		Deletes all data from the file which resides to the left of the cursor. For example,
 *
 * 			This is example d|ata
 * 			<br>in some file.
 *
 * 		where "|" is the cursor. After delete is called, this will become,
 *
 * 			|ata
 * 			<br>in some file.
 *
 * 		The cursor will be at the beginning of the file after the method exits.
 *
 * 		<b>Parameter</b>
 * 			<br>sd: A pointer to the mission_logger_t structure used to access this method.
 *
 * 		<b>Returns</b>
 *			<br>MISSION_LOGGER_OK		if successful
 *			<br>MISSION_LOGGER_FAILURE	if the method failed. The cursor position is garbage. No data was
 *			erased
 *			<br>MISSION_LOGGER_NO_FILE	No file was open. Nothing was done.
 *
 * @var mission_logger_t::reformat
 * 		A function pointer with prototype,
 * 			@code
 * 			ml_result_t reformat( mission_logger_t *sd );
 * 			@endcode
 * 		Completely reformats the file system being used by the logger. This is an emergency method only.
 *
 * 		<b>Parameters</b>
 * 			<br>sd: A pointer to the mission_logger_t structure used to access this method
 *
 * 		<b>Returns</b>
 *			<br>MISSION_LOGGER_OK		if successful
 *			<br>MISSION_LOGGER_FAILURE	if the method failed. The file system was unchanged.
 */
struct mission_logger_t
{
	void *__data;
	const char *__file;
	ml_result_t (*open_a)( mission_logger_t *, const char * );
	ml_result_t (*open_r)( mission_logger_t *, const char * );
	ml_result_t (*close)( mission_logger_t * );
	ml_result_t (*file_size)( mission_logger_t *, unsigned long * );
	ml_result_t (*write)( mission_logger_t *, const char *, unsigned int, unsigned int * );
	ml_result_t (*write_stream)( mission_logger_t *, const uint8_t *, unsigned int, unsigned int * );
	ml_result_t (*read)( mission_logger_t *, const char *, unsigned int, unsigned int * );
	ml_result_t (*go_to)( mission_logger_t *, unsigned int );
	ml_result_t (*delete)( mission_logger_t * );
	ml_result_t (*reformat)( mission_logger_t * );
};

/**
 * @brief
 * 		Reads data out a file and backs it up.
 *
 * @details
 * 		Reads data out a file and backs it up. Data will be read starting at the
 * 		beginning of the file and continue for the number of bytes specified. The
 * 		read data deleted from the file and backed up into the specified back up
 * 		file.
 *
 * @param logger
 * 		A pointer to the mission_logger_t structure to use for the operation.
 *
 * @param msg
 * 		A pointer to the array the read message will be put in.
 *
 * @param size
 * 		The size of the msg array. This is also the number of bytes to read and
 * 		backup.
 *
 * @param file
 * 		The file to read data out of.
 *
 * @param backup
 * 		The file to backup data in.
 *
 * @returns
 * 		MISSION_LOGGER_OK if successful
 * 		<br>MISSION_LOGGER_FAILURE if unsuccessful, no data is lost or corrupt.
 */
ml_result_t read_and_backup( mission_logger_t *logger, char *msg, unsigned int size,
							 char const *file, char const *backup );

/**
 * @brief
 * 		Read hk data of eps in FIFO style and back it up.
 *
 * @details
 * 		Read hk data of eps in FIFO style and back it up. Data that is read
 * 		gets deleted (from where it was read) and backed up in a log file.
 *
 * @param logger
 * 		A pointer to the mission_logger_t sruct to use for the operation.
 *
 * @param msg
 * 		A pointer to the array to store the read data.
 *
 * @param size
 * 		The length of the msg array and the number of bytes that will be read.
 *
 * @returns
 * 		MISSION_LOGGER_OK		if successful
 * 		<br>MISSION_LOGGER_FAILRUE	if unsuccessful. No data gets corrupted.
 */
ml_result_t read_and_backup_eps_hk( mission_logger_t *logger, char *msg, unsigned int size );

/**
 * @brief
 * 		Read hk data of adcs in FIFO style and back it up.
 *
 * @details
 * 		Read hk data of adcs in FIFO style and back it up. Data that is read
 * 		gets deleted (from where it was read) and backed up in a log file.
 *
 * @param logger
 * 		A pointer to the mission_logger_t sruct to use for the operation.
 *
 * @param msg
 * 		A pointer to the array to store the read data.
 *
 * @param size
 * 		The length of the msg array and the number of bytes that will be read.
 *
 * @returns
 * 		MISSION_LOGGER_OK		if successful
 * 		<br>MISSION_LOGGER_FAILRUE	if unsuccessful. No data gets corrupted.
 */
ml_result_t read_and_backup_adcs_hk( mission_logger_t *logger, char *msg, unsigned int size );

/**
 * @brief
 * 		Read hk data of comm in FIFO style and back it up.
 *
 * @details
 * 		Read hk data of comm in FIFO style and back it up. Data that is read
 * 		gets deleted (from where it was read) and backed up in a log file.
 *
 * @param logger
 * 		A pointer to the mission_logger_t sruct to use for the operation.
 *
 * @param msg
 * 		A pointer to the array to store the read data.
 *
 * @param size
 * 		The length of the msg array and the number of bytes that will be read.
 *
 * @returns
 * 		MISSION_LOGGER_OK		if successful
 * 		<br>MISSION_LOGGER_FAILRUE	if unsuccessful. No data gets corrupted.
 */
ml_result_t read_and_backup_comm_hk( mission_logger_t *logger, char *msg, unsigned int size );

/**
 * @brief
 * 		Write eps hk data to memory.
 *
 * @details
 * 		Write eps hk data to memory. Data is written as a byte stream.
 *
 * @param logger
 * 		A pointer to the mission_logger_t structure to use when writting data.
 *
 * @param msg
 * 		A pointer to the data to write.
 *
 * @param size
 * 		The length of the msg array. The number of bytes to write.
 *
 * @returns
 * 		MISSION_LOGGER_OK			if successful
 * 		<br>MISSION_LOGGER_FAILURE	in unsuccessful. No data is written.
 */
ml_result_t write_eps_hk( mission_logger_t *logger, char const *msg, unsigned int size );

/**
 * @brief
 * 		Write adcs hk data to memory.
 *
 * @details
 * 		Write adcs hk data to memory. Data is written as a byte stream.
 *
 * @param logger
 * 		A pointer to the mission_logger_t structure to use when writting data.
 *
 * @param msg
 * 		A pointer to the data to write.
 *
 * @param size
 * 		The length of the msg array. The number of bytes to write.
 *
 * @returns
 * 		MISSION_LOGGER_OK			if successful
 * 		<br>MISSION_LOGGER_FAILURE	in unsuccessful. No data is written.
 */
ml_result_t write_adcs_hk( mission_logger_t *logger, char const* msg, unsigned int size );

/**
 * @brief
 * 		Write comm hk data to memory.
 *
 * @details
 * 		Write comm hk data to memory. Data is written as a byte stream.
 *
 * @param logger
 * 		A pointer to the mission_logger_t structure to use when writting data.
 *
 * @param msg
 * 		A pointer to the data to write.
 *
 * @param size
 * 		The length of the msg array. The number of bytes to write.
 *
 * @returns
 * 		MISSION_LOGGER_OK			if successful
 * 		<br>MISSION_LOGGER_FAILURE	in unsuccessful. No data is written.
 */
ml_result_t write_comm_hk( mission_logger_t *logger, char const *msg, unsigned int size );

/* Temporary methods for retrieving housekeeping. These are for use in ECE
 * 492 demo only. */
unsigned int next_comm_hk( mission_logger_t *, char *, unsigned int );
unsigned int next_eps_hk( mission_logger_t *, char *, unsigned int );
unsigned int next_adcs_hk( mission_logger_t *, char *, unsigned int );
unsigned int next_state_hk( mission_logger_t *, char *, unsigned int );
void log_state_change( mission_logger_t *, uint8_t );

/**
 * @brief
 * 		Initializes a mission_logger_t structure to use an sd card as a memory
 * 		resource and the fatfs file system for memory management on an lpc1769
 * 		board.
 *
 * @details
 * 		Initializes a mission_logger_t structure to use an sd card as a memory
 * 		resource and the fatfs file system for memory management on an lpc1769
 * 		board. The external hardware is assumed to be an lpc1769 with spi channel
 * 		zero being connected to the sd card and pin 18 being the chip
 * 		select pin.
 *
 * @attention
 * 		This method must not be used before the FreeRTOS scheduler is started. Using
 * 		this method before the scheduler starts will result in undefined
 * 		behaviour.
 *
 * @attention
 * 		Calls to vTaskDelay and vTaskDelayUntil must NOT be made while a file
 * 		is open. This will result in a hard fault.
 *
 * @remark
 * 		This initialization makes the mission_logger_t struct part of a singleton.
 * 		All mission loggers initialized this way share resources using the same mutexs.
 *
 * @param sd
 * 		A pointer to the mission_logger_t structure to initialize.
 *
 * @returns
 * 		MISSION_LOGGER_OK 			if initialization succeeds
 * 		<br>MISSION_LOGGER_FAILURE	if initialization failed.
 */
ml_result_t initialize_mission_logger_lpc_fatfs_sd( mission_logger_t *sd );

#endif /* INCLUDE_MISSION_LOGGER_H_ */
