/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file TelecommandPacket.h
 * @author Brendan Bruner
 * @date 2015-01-12
 *
 * Defines the packet format for a telecommand from gound station.
 */

#ifndef TELECOMMANDPACKET_H_
#define TELECOMMANDPACKET_H_

#define PRIORITY_MASK	0xfe
#define	OPTIONS_MASK	0x01

/**
 * @struct telecommand_packet_t
 *
 * @brief
 * 		Packet format of telecommands.
 *
 * @details
 *		Packet format of telecommands.
 *
 * @var telecommand_packet_t::priority_and_option
 * 		 Bit one indicates if there is an additional payload field. If bit
 * 		 one is set, there is additional payload data. If bit one is unset
 * 		 there is no additional payload data.
 *
 * @var telecommand_packet_t::payload_bytes
 * 		The number of bytes of payload data IF bit one of priority_and_option
 * 		was set.
 *
 * @var telecommand_packet_t::payload_location
 * 		A pointer to the location of payload data. The array pointed to is
 * 		payload_bytes long.
 *
 * @var telecommand_packet_t::time_of_creation
 * 		UTC time of when the telecommand was created.
 *
 * @var telecommand_packet_t::time_of_execution
 * 		The number of milliseconds to wait before the command is executed.
 * 		The measurement is relative to time_of_creation. Therefore, a value
 * 		of zero means immediate execution.
 *
 * @var telecommand_packet_t::telecommand_code
 * 		The id of the telecommand. This field tells the OBC what command
 * 		this packet refers to. For example, a value of zero might indicate
 * 		a soft reset command should be executed.
 *
 * @var telecommand_packet_t::unique_id
 * 		The unique id of the telecommand and hence the command which will
 * 		be executed. No telecommands have the same id, even if their
 * 		telecommand_code is the same.
 */
typedef struct __attribute__((packed))
{
	uint8_t 	priority_and_option;
	uint8_t 	payload_bytes;
	uint32_t 	time_of_creation;
	uint32_t 	ms_to_execution;
	uint16_t	telecommand_code;
	uint32_t	unique_id;
	uint8_t		payload_location;
} telecommand_packet_t;

#define extract_priority( tc )			( ((tc)->priority_and_option & PRIORITY_MASK) >> 1 )
#define extract_options( tc ) 			( (tc)->priority_and_option & OPTIONS_MASK )
#define extract_payload_size( tc )		( (tc)->payload_bytes )
#define extract_payload_location( tc )	( (tc)->payload_location )
#define extract_creation_time( tc )		( (tc)->time_of_creation )
#define extract_execution_time( tc )	( (tc)->ms_to_execution )
#define extract_code( tc )				( (tc)->telecommand_code )
#define extract_id( tc )				( (tc)->unique_id )

#endif /* TELECOMMANDPACKET_H_ */
