/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file comm_arduino_remote
 * @author Brendan Bruner
 * @date Mar 30, 2015
 */
#include "comm.h"
#include "uart.h"
#include "FreeRTOS.h"
#include "../../Core/include/task_defines.h"

/************************************************************************/
/* PRIVATE DEFINES														*/
/************************************************************************/
#define INITIALIZED		1
#define NOT_INITIALIZED	0

#define ONE_BYTE		1

#define NULL_MEMORY		0
#define NON_NULL_MEMORY	1

/************************************************************************/
/* PRIVATE GLOBALS														*/
/************************************************************************/
static uint8_t is_initialized = NOT_INITIALIZED;

/************************************************************************/
/* PRIVATE METHODS														*/
/************************************************************************/
static uint8_t null_memory( uint8_t *data, uint32_t size )
{
	for( ; size > 0; --size )
	{
		if( data[size-1] != 0 )
		{
			return NON_NULL_MEMORY;
		}
	}
	return NULL_MEMORY;
}


/************************************************************************/
/* VIRTUAL METHOD IMPLEMENTATIONS										*/
/************************************************************************/
static comm_beacon_t beacon_mock( comm_t *comm )
{
	uint8_t msg[] = {0};
	uint8_t read = 0;

	serial2_lock( );
	serial2_reset( );
	serial2_send( msg, 1 );
	while( read < 1 )
	{
		read += serial2_read( (uint8_t *) (msg+read), 1-read );
	}
	serial2_unlock( );

	return (comm_beacon_t) msg[0];
}

static int com_set_conf_mock( comm_t *comm, nanocom_conf_t *config )
{
	return 0;
}

static int com_get_conf_mock( comm_t *comm, nanocom_conf_t *config )
{
	return 0;
}

static int com_restore_conf_mock( comm_t *comm )
{
	return 0;
}

static int com_get_status_mock(comm_t *comm, nanocom_data_t *status )
{
	return 0;
}

static int com_get_log_rssi_mock(comm_t *comm, nanocom_rssi_t *rssi, uint8_t *count )
{
	return 0;
}

static int com_get_hk_mock(comm_t *comm, nanocom_hk_t *hk, uint8_t *count )
{
	uint8_t const msg[] = {1};
	uint32_t read = 0;
	uint8_t *bit_stream = (uint8_t *) hk;

	serial2_lock( );
	serial2_reset( );
	serial2_send( msg, 1 );
	while( read < HK_ARRAY_SIZE * sizeof( nanocom_hk_t ) )
	{
//		bit_stream += read;
		read += serial2_read( (uint8_t *) (bit_stream+read), (HK_ARRAY_SIZE * sizeof( nanocom_hk_t )) - read );
	}
	serial2_unlock( );

	*count = HK_ARRAY_SIZE;

	return 0;
}

static uint8_t poll_telecommand( comm_t *comm, telecommand_packet_t *command )
{
	uint8_t msg_out[1] = {2};
	uint8_t *msg_in;
	uint32_t read;

	read = 0;
	msg_in = (uint8_t *) command;
	if( comm->beacon( comm ) == COMM_IS_BEACONED )
	{
		serial2_lock( );
		serial2_reset( );
		serial2_send( msg_out, ONE_BYTE );
		serial2_reset( );
		while( read < sizeof( telecommand_packet_t ) )
		{
			read += serial2_read( (uint8_t *) (msg_in+read), sizeof( telecommand_packet_t ) - read );
		}
		serial2_unlock( );
		if( null_memory( msg_in, sizeof( telecommand_packet_t ) ) == NULL_MEMORY )
		{
			return NO_TELECOMMAND;
		}
		return INCOMING_TELECOMMAND;
	}
	return NO_TELECOMMAND;
}

static uint8_t transmit( comm_t *comm, uint8_t *data, uint16_t size )
{
	uint8_t msg_out[1] = {5};
	uint8_t msg[1];
	uint32_t written, read;

	written = 0;
	read = 0;
	if( comm->beacon( comm ) == COMM_IS_BEACONED )
	{
		serial2_lock( );
		serial2_reset( );
		serial2_send( msg_out, ONE_BYTE );
		serial2_reset( );
		written = serial2_send( data, (uint8_t) size );
		while( read < 1 )
		{
			read += serial2_read( (uint8_t *) (msg+read), 1-read );
		}
		serial2_unlock( );
		return written;
	}
	return 0;
}

/************************************************************************/
/* PRIVATE METHODS														*/
/************************************************************************/
static void comm_gateway( void *args )
{
	for( ;; )
	{
		vTaskDelay( COMM_GATEWAY_PERIOD );
	}
}

/************************************************************************/
/* Initialization 														*/
/************************************************************************/
void initialize_comm_arduino_remote( comm_t *comm )
{
	serial2_begin( );

	comm->beacon = &beacon_mock;
	comm->com_set_conf = &com_set_conf_mock;
	comm->com_get_conf = &com_get_conf_mock;
	comm->com_restore_conf = &com_restore_conf_mock;
	comm->com_get_status = &com_get_status_mock;
	comm->com_get_log_rssi = &com_get_log_rssi_mock;
	comm->com_get_hk = &com_get_hk_mock;
	comm->poll_telecommand = &poll_telecommand;
	comm->transmit = &transmit;

	if( is_initialized == NOT_INITIALIZED )
	{
		xTaskCreate( comm_gateway, COMM_GATEWAY_NAME,
					 COMM_GATEWAY_STACK, NO_PARAMETERS,
					 COMM_GATEWAY_PRIO, NO_HANDLE );
		is_initialized = INITIALIZED;
		/* TODO: handle failed task creation. */
	}
}
