/* FreeRTOS.org includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"

#include "i2c.h"
#include "adcs.h"
#include <stdlib.h>

#define I2C_MASTER  (LPC_I2C1)
/* Global variables to keep track of transfers */
__IO FlagStatus complete_M;

static SemaphoreHandle_t i2c_sem;

/* Transmit and receive buffers for Master and Slave */
//char Master_Buf[BUFFER_SIZE];

/*********************************************************************//**
 * @brief		Interrupt Service Routine for Master Peripheral
 * 				Assigns Handle to interrupt handler and sets flag
 * 				upon completion
 * @param[in]	None
 * @return		None
 **********************************************************************/
void I2C1_IRQHandler() {
	I2C_MasterHandler(I2C_MASTER);
	if(I2C_MasterTransferComplete(I2C_MASTER)) {
		complete_M = SET;
	}
}


/*********************************************************************//**
 * @brief		Initialize Master buffer with 0s and Slave buffer with 1s
 * @param[in]	None
 * @return		None
 **********************************************************************/
void buffer_init(uint8_t* Master_Buf, int rx_length) {
	int i;
	for(i = 0; i<rx_length; i++) {
			*(Master_Buf+i) = 0;
	}
}
/*********************************************************************//**
 * @brief		I2C initialization function
 * @param[in]	None
 * @return		None
 **********************************************************************/
void setupI2C ()
{
	PINSEL_CFG_Type PinCfg;

	i2c_sem =  xSemaphoreCreateMutex();

	/* Init I2C pins */
	PinCfg.Portnum   = 0;
	PinCfg.Pinnum    = 0;
	PinCfg.Funcnum   = PINSEL_FUNC_3;
	PinCfg.OpenDrain = PINSEL_PINMODE_NORMAL;
	PinCfg.Pinmode   = PINSEL_PINMODE_PULLUP;
	PINSEL_ConfigPin(&PinCfg);
	PinCfg.Pinnum    = 1;
	PINSEL_ConfigPin(&PinCfg);


	/* Initialize Master I2C peripheral */
	I2C_Init(I2C_MASTER, I2C_CLK_RATE);

	/* Set up interrupts
	 * Note: Master has higher priority
	 */
	NVIC_DisableIRQ(I2C1_IRQn);
	NVIC_SetPriority(I2C1_IRQn, ((0x00<<3)|0x01));

	/*Enable Master and Slave I2C peripherals */
	I2C_Cmd(I2C_MASTER, ENABLE);

}

void i2c_lock( void )
{
	xSemaphoreTake( i2c_sem, portMAX_DELAY );
}

void i2c_unlock(void)
{
	xSemaphoreGive( i2c_sem );
}

void* I2CRecieve(I2C_M_SETUP_Type MasterConfig, uint32_t temp, int rx_length)
{
	/* MASTER SEND DATA TO SLAVE -------------------------------------------------------- */
	/* Force complete flag for the first time of running */
	complete_M =  RESET;
	uint8_t Master_Buf;
	//uint8_t* Master_Buf = (uint8_t*) malloc(rx_length);
	buffer_init(&Master_Buf, rx_length);
	//printf("Master receiving from slave...\n");

	/* Start I2C master device*/
	MasterConfig.sl_addr7bit = I2C_SLAVE_ADDR;
	MasterConfig.tx_data = NULL;
	MasterConfig.tx_length = 0;
	MasterConfig.rx_data = &Master_Buf;
	MasterConfig.rx_length = rx_length;
	MasterConfig.retransmissions_max = 3;
	I2C_MasterTransferData(I2C_MASTER, &MasterConfig, I2C_TRANSFER_INTERRUPT);

	/* Wait until both of them complete */
	while ((complete_M == RESET));

	/*printf("Master buffer: ");

	adcs_unix_time_t* time = (adcs_unix_time_t*)Master_Buf;
	printf("unix_time: %d\n", time->unix_time);
	printf("millis: %d\n", time->millis);
	//printf("error flags: %d\n", comm_status->comm_status_booleans);
	printf("\n");
	free(Master_Buf);*/
	return (void*) Master_Buf;
}

void I2CSend(I2C_M_SETUP_Type MasterConfig, uint32_t temp, uint8_t cmd_id, uint8_t* input_struct, uint8_t size_of_struct)
{

	/* MASTER SEND DATA TO SLAVE -------------------------------------------------------- */
	/* Force complete flag for the first time of running */
	complete_M =  RESET;

	uint8_t send_struct;
	//uint8_t* send_struct = (uint8_t*) malloc(size_of_struct+1);

	send_struct = cmd_id;
	/*int i;
	for (i = 1; i<size_of_struct+1; i++)
	{
		*(send_struct + i) = *(input_struct+i-1);
	}*/

	/* Start I2C master device*/
	MasterConfig.sl_addr7bit = I2C_SLAVE_ADDR;
	MasterConfig.rx_data = NULL;
	MasterConfig.rx_length = 0;
	MasterConfig.tx_data = &send_struct;//&cmd_id;
	MasterConfig.tx_length = size_of_struct+1;
	MasterConfig.retransmissions_max = 3;
	I2C_MasterTransferData(I2C_MASTER, &MasterConfig, I2C_TRANSFER_INTERRUPT);

	/* Wait until both of them complete */
	while ((complete_M == RESET));

	/*if (cmd_id < 127)
	{
		complete_M =  RESET;

		 Start I2C master device
		MasterConfig.sl_addr7bit = I2C_SLAVE_ADDR;
		MasterConfig.rx_data = NULL;
		MasterConfig.rx_length = 0;
		MasterConfig.tx_data = input_struct;
		MasterConfig.tx_length = size_of_struct;
		MasterConfig.retransmissions_max = 3;
		I2C_MasterTransferData(I2C_MASTER, &MasterConfig, I2C_TRANSFER_INTERRUPT);

		 Wait until both of them complete
		while ((complete_M == RESET));
	}*/
}

/*********************************************************************//**
 * @brief		Task to test Master-Slave Communication
 * @param[in]	None
 * @return		None
 **********************************************************************/
void I2C_test ( void *pvParameters )
{
	I2C_M_SETUP_Type MasterConfig;
	uint32_t temp;

	adcs_unix_time_t time;
	time.unix_time = 987654321;
	time.millis = 54321;
	temp = 0;

	/* this loop prevents the task from ending */
	for ( ; ; ) {
		I2CSend(MasterConfig, temp, 2, (uint8_t *) &time, sizeof(adcs_unix_time_t));
		vTaskDelay ( 1000 / portTICK_RATE_MS );
		I2CSend(MasterConfig, temp, 143, NULL, 0);
		//vTaskDelay ( 1000 / portTICK_RATE_MS );
		I2CRecieve(MasterConfig, temp, 6);
		vTaskDelay ( 1000 / portTICK_RATE_MS );
	}
}

