/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file mission_logger
 * @author Brendan Bruner
 * @date Mar 26, 2015
 */

#include "mission_logger.h"

#define HK_EPS_FILE		"eps_hk.txt"
#define HK_EPS_BACKUP	"eps_backup.txt"
#define HK_ADCS_FILE	"adcs_hk.txt"
#define HK_ADCS_BACKUP	"adcs_backup.txt"
#define HK_COMM_FILE	"comm_hk.txt"
#define HK_COMM_BACKUP	"comm_backup.txt"
#define HK_STATE_FILE	"state_hk.txt"

#define NEW_LINE		"\n"
#define NEW_LINE_SIZE	1

#define ASSERT_LOGGER_SUCCESS( lr, log )				\
				do {									\
					if( lr != MISSION_LOGGER_OK )		\
					{									\
						log->close( log );				\
						return MISSION_LOGGER_FAILURE;	\
					}									\
				} while( 0 )


ml_result_t read_and_backup( mission_logger_t *logger, char *msg, unsigned int size,
							 char const *file, char const *backup )
{
	unsigned int read, written;
	ml_result_t mlr;

	/* Close any open files */
	logger->close( logger );

	mlr = logger->open_a( logger, file );
	ASSERT_LOGGER_SUCCESS( mlr, logger );
	mlr = logger->go_to( logger, BEGINNING_OF_FILE );
	ASSERT_LOGGER_SUCCESS( mlr, logger );
	mlr = logger->read( logger, msg, size, &read);
	if( mlr != MISSION_LOGGER_END_OF_FILE )
	{
		ASSERT_LOGGER_SUCCESS( mlr, logger );
	}
	if( read > 0 )
	{
		//mlr = logger->delete( logger );
		//ASSERT_LOGGER_SUCCESS( mlr, logger );
		mlr = logger->close( logger );
		//ASSERT_LOGGER_SUCCESS( mlr, logger );

		/* TODO: better fault tolerance than a do while */
		do {
			mlr = logger->open_a( logger, backup );
		} while( mlr != MISSION_LOGGER_OK );
		do {
			mlr = logger->write( logger, msg, read, &written );
		} while( mlr != MISSION_LOGGER_OK );
		mlr = logger->close( logger );
		ASSERT_LOGGER_SUCCESS( mlr, logger );
	}
	else
	{
		mlr = logger->close( logger );
		ASSERT_LOGGER_SUCCESS( mlr, logger );
	}
	return MISSION_LOGGER_OK;
}

ml_result_t read_and_backup_eps_hk( mission_logger_t *logger, char *msg, unsigned int size )
{
	return read_and_backup( logger, msg, size, HK_EPS_FILE, HK_EPS_BACKUP );
}

ml_result_t read_and_backup_adcs_hk( mission_logger_t *logger, char *msg, unsigned int size )
{
	return read_and_backup( logger, msg, size, HK_ADCS_FILE, HK_ADCS_BACKUP );
}

ml_result_t read_and_backup_comm_hk( mission_logger_t *logger, char *msg, unsigned int size )
{
	return read_and_backup( logger, msg, size, HK_COMM_FILE, HK_COMM_BACKUP );
}

ml_result_t write_eps_hk( mission_logger_t *logger, char const *msg, unsigned int size )
{
	unsigned int written;
	ml_result_t mlr;

	/* Close any open files. */
	logger->close( logger );

	mlr = logger->open_a( logger, HK_EPS_FILE );
	ASSERT_LOGGER_SUCCESS( mlr, logger );
	mlr = logger->write( logger, msg , size, &written );
	ASSERT_LOGGER_SUCCESS( mlr, logger );
	logger->close( logger );
	return MISSION_LOGGER_OK;
}

ml_result_t write_adcs_hk( mission_logger_t *logger, char const *msg, unsigned int size )
{
	unsigned int written;
	ml_result_t mlr;

	/* Close any open files. */
	logger->close( logger );

	mlr = logger->open_a( logger, HK_ADCS_FILE );
	ASSERT_LOGGER_SUCCESS( mlr, logger );
	mlr = logger->write( logger, msg , size, &written );
	ASSERT_LOGGER_SUCCESS( mlr, logger );
	logger->close( logger );
	return MISSION_LOGGER_OK;
}

ml_result_t write_comm_hk( mission_logger_t *logger, char const *msg, unsigned int size )
{
	unsigned int written;
	ml_result_t mlr;

	/* Close any open files. */
	logger->close( logger );

	mlr = logger->open_a( logger, HK_COMM_FILE );
	ASSERT_LOGGER_SUCCESS( mlr, logger );
	mlr = logger->write( logger, msg , size, &written );
	ASSERT_LOGGER_SUCCESS( mlr, logger);
	mlr = logger->write( logger, NEW_LINE, NEW_LINE_SIZE, &written );
	logger->close( logger );
	return MISSION_LOGGER_OK;
}

unsigned int next_comm_hk( mission_logger_t *logger, char *msg, unsigned int size )
{
	static unsigned long next_position = 0;
	ml_result_t mlr;
	unsigned int read;

	/* Close any open files. */
	logger->close( logger );

	mlr = logger->open_a( logger, HK_COMM_FILE );
	if( mlr != MISSION_LOGGER_OK )
	{
		logger->close( logger );
		return 0;
	}
	mlr = logger->go_to( logger, next_position );
	if( mlr != MISSION_LOGGER_OK )
	{
		logger->close( logger );
		return 0;
	}

	mlr = logger->read( logger, msg, size, &read);
	logger->close( logger );
	if( mlr != MISSION_LOGGER_END_OF_FILE && mlr != MISSION_LOGGER_OK )
	{
		return 0;
	}
	if( read == size )
	{
		next_position += read;
		return read;
	}
	return 0;
}

unsigned int next_eps_hk( mission_logger_t *logger, char *msg, unsigned int size )
{
	static unsigned int next_position = 0;
	ml_result_t mlr;
	unsigned int read;

	/* Close any open files. */
	logger->close( logger );

	mlr = logger->open_a( logger, HK_EPS_FILE );
	if( mlr != MISSION_LOGGER_OK )
	{
		logger->close( logger );
		return 0;
	}
	mlr = logger->go_to( logger, next_position );
	if( mlr != MISSION_LOGGER_OK )
	{
		logger->close( logger );
		return 0;
	}

	mlr = logger->read( logger, msg, size, &read);
	logger->close( logger );
	if( mlr != MISSION_LOGGER_END_OF_FILE && mlr != MISSION_LOGGER_OK )
	{
		return 0;
	}
	if( read == size )
	{
		next_position += read;
		return read;
	}
	return 0;
}

unsigned int next_adcs_hk( mission_logger_t *logger, char *msg, unsigned int size )
{
	static unsigned int next_position = 0;
	ml_result_t mlr;
	unsigned int read;

	/* Close any open files. */
	logger->close( logger );

	mlr = logger->open_a( logger, HK_ADCS_FILE );
	if( mlr != MISSION_LOGGER_OK )
	{
		logger->close( logger );
		return 0;
	}
	mlr = logger->go_to( logger, next_position );
	if( mlr != MISSION_LOGGER_OK )
	{
		logger->close( logger );
		return 0;
	}

	mlr = logger->read( logger, msg, size, &read);
	logger->close( logger );
	if( mlr != MISSION_LOGGER_END_OF_FILE && mlr != MISSION_LOGGER_OK )
	{
		return 0;
	}
	if( read == size )
	{
		next_position += read;
		return read;
	}
	return 0;
}

unsigned int next_state_hk( mission_logger_t *logger, char *msg, unsigned int size )
{
	static unsigned int next_position = 0;
	ml_result_t mlr;
	unsigned int read;

	/* Close any open files. */
	logger->close( logger );

	mlr = logger->open_a( logger, HK_STATE_FILE );
	if( mlr != MISSION_LOGGER_OK )
	{
		logger->close( logger );
		return 0;
	}
	mlr = logger->go_to( logger, next_position );
	if( mlr != MISSION_LOGGER_OK )
	{
		logger->close( logger );
		return 0;
	}

	mlr = logger->read( logger, msg, size, &read);
	logger->close( logger );
	if( mlr != MISSION_LOGGER_END_OF_FILE && mlr != MISSION_LOGGER_OK )
	{
		return 0;
	}
	if( read == size )
	{
		next_position += read;
		return read;
	}
	else
	{
		return 0;
	}
}

void log_state_change( mission_logger_t *logger, uint8_t state )
{
	unsigned int written;
	ml_result_t mlr;

	/* Close any open files. */
	logger->close( logger );

	mlr = logger->open_a( logger, HK_STATE_FILE );
	if( mlr != MISSION_LOGGER_OK )
	{
		logger->close( logger );
		return;
	}
	mlr = logger->write( logger, (char *) &state, 1, &written );
	logger->close( logger );
}
