/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 */
/**
 * @file hk_logger_fatfs_sd
 * @author Brendan Bruner
 * @date Mar 16, 2015
 */

#include "mission_logger.h"
#include "spi_sd_lpc17xx.h"
#include "FreeRTOS.h"
#include "task.h"
#include "core_cm3.h"
#include "lpc17xx_timer.h"
#include <stdio.h>

static uint8_t byte_to_ascii( uint8_t byte, char *str );
/************************************************************************/
/* Private Defines														*/
/************************************************************************/
/* Print debug info to console when defined */
//#define MISSION_LOGGER_DEBUG

#define MISSION_LOGGER_RETRY_MAX	(6)

#define NO_FILE 		(char const *) 0
#define DEFAULT_DRIVE	""
#define MOUNT_LATER		0

#define COPY_BUFFER_SIZE (128)
#define COPY_BUFFER_EXP	(7)
#define DELETE_BUFFER	("__delete_buffer.txt")

#define ML_UNINITIALIZED	((unsigned char) 0)
#define ML_INITIALIZED		((unsigned  char) 1)

#define BEGINNING_OF_FILE	0

#define ASSERT_SUCCESS( fr, file_obj )							\
						do {									\
							if( fr != FR_OK )					\
							{									\
								f_close( file_obj );			\
								return MISSION_LOGGER_FAILURE;	\
							}									\
						} while( 0 )
#define ASSERT_FR( fr )											\
						do {									\
							if( fr != FR_OK )					\
							{									\
								return MISSION_LOGGER_FAILURE;	\
							}									\
						} while( 0 )
#define ASSERT_OPEN( sd )												\
					do {												\
						if( sd.__name == NO_FILE )						\
						{												\
							return MISSION_LOGGER_NO_FILE; 				\
						}												\
					} while( 0 )

#define MISSION_LOGGER_TRY												\
					do {												\
						int __try_iter;									\
						for( __try_iter = 0; __try_iter < MISSION_LOGGER_RETRY_MAX; ++__try_iter )	\
						{
#define MISSION_LOGGER_EXIT_TRY( fr, file_obj_p )						\
							if( fr == FR_OK )							\
							{											\
								break;									\
							}											\
							f_close( file_obj_p );						\
						}												\
					} while( 0 )


/************************************************************************/
/* Private Variable Types												*/
/************************************************************************/
/*
 * structure: fatfs_t
 *
 * __fatfs:
 * 		Work area (file system object) for logical drive
 *
 * __file:
 * 		File object
 *
 * __open_mutex, __write_mutex:
 * 		Mutexs to provide mutual exclusion when opening and writing /
 * 		reading from files
 *
 * __is_init:
 * 		Prevents duplicate initializations
 *
 * __name:
 * 		name of the open file.
 */
typedef struct
{
	FATFS 				__fatfs;
	FIL 				__file;
	FIL					__delete_buffer;
	SemaphoreHandle_t 	__open_mutex;
	SemaphoreHandle_t	__write_mutex;
	SemaphoreHandle_t	__delete_mutex;
	unsigned char 		__is_init;
	char const 			*__name;
} fatfs_t;

static fatfs_t file_system = { .__is_init = ML_UNINITIALIZED,
							   .__name = NO_FILE 				};

/************************************************************************/
/* Virtual Methods														*/
/************************************************************************/
static ml_result_t open_a( mission_logger_t *sd, const char *file_name )
{
	FRESULT fr;

	/* Block until the currently open file is closed. */
	xSemaphoreTake( file_system.__open_mutex, portMAX_DELAY );

	/* This must be done before any ASSERTS to prevent deadlock. */
	file_system.__name = file_name;

	/* Enter a try for opening, and one for seeking. */
	MISSION_LOGGER_TRY
	MISSION_LOGGER_TRY
	/* Open the file. */
	fr = f_open( &file_system.__file, file_name, FA_READ | FA_WRITE | FA_OPEN_ALWAYS );
	MISSION_LOGGER_EXIT_TRY( fr, &file_system.__file );

#ifdef MISSION_LOGGER_DEBUG
	printf( "f_open, open_a: %d\n", fr );
#endif

	/* Seek to the end of the file. */
	fr = f_lseek( &file_system.__file, f_size( &file_system.__file ) );
	MISSION_LOGGER_EXIT_TRY( fr, &file_system.__file );

#ifdef MISSION_LOGGER_DEBUG
	printf( "f_lseek, open_a: %d\n", fr );
#endif

	ASSERT_SUCCESS( fr, &file_system.__file );
	return MISSION_LOGGER_OK;
}

static ml_result_t open_r( mission_logger_t *sd, const char *file_name )
{
	FRESULT fr;

	/* Block until the currently open file is closed. */
	xSemaphoreTake( file_system.__open_mutex, portMAX_DELAY );

	/* Enter a mission logger try for opening and seeking. */
	MISSION_LOGGER_TRY
	MISSION_LOGGER_TRY

	/* Open the file. */
	fr = f_open( &file_system.__file, file_name, FA_READ | FA_OPEN_ALWAYS );
	MISSION_LOGGER_EXIT_TRY( fr, &file_system.__file );

#ifdef MISSION_LOGGER_DEBUG
	printf( "f_open, open_r: %d\n", fr );
#endif

	/* Seek to the beginning of the file. */
	file_system.__name = file_name;
	fr = f_lseek( &file_system.__file, BEGINNING_OF_FILE );
	MISSION_LOGGER_EXIT_TRY( fr, &file_system.__file );

#ifdef MISSION_LOGGER_DEBUG
	printf( "f_lseek, open_r: %d\n", fr );
#endif

	ASSERT_SUCCESS( fr, &file_system.__file );
	return MISSION_LOGGER_OK;
}

static ml_result_t close( mission_logger_t *sd )
{
	/* Assert there is an open file to close. */
	ASSERT_OPEN( file_system );

	/* Close the currently open file. */
	f_close(  &file_system.__file );
	file_system.__name = NO_FILE;

#ifdef MISSION_LOGGER_DEBUG
	printf( "f_close, close: %d\n", fr );
#endif

	/* Give back the mutex protecting the open file. */
	xSemaphoreGive( file_system.__open_mutex );
	return MISSION_LOGGER_OK;
}

static ml_result_t file_size( mission_logger_t *sd, unsigned long *size )
{
	/* Assert there is an open file to query. */
	ASSERT_OPEN( file_system );

	/* Mutual exclusion on file io */
	xSemaphoreTake( file_system.__write_mutex, portMAX_DELAY );
	*size = f_size( &file_system.__file );
	xSemaphoreGive( file_system.__write_mutex );

	return MISSION_LOGGER_OK;
}

static ml_result_t write( mission_logger_t *sd, const char *msg, unsigned int size, unsigned int *written )
{
	FRESULT	fr;

	/* Assert there is an open file. */
	ASSERT_OPEN( file_system );

	/* Put the disk io under mutual exclusion. */
	xSemaphoreTake( file_system.__write_mutex, portMAX_DELAY );
	fr = f_write( &file_system.__file, msg, size, written );
#ifdef MISSION_LOGGER_DEBUG
	printf( "f_write, write: %d\n", fr );
#endif

	/* Return semaphore so other processes can write / read from file. */
	xSemaphoreGive( file_system.__write_mutex );

	ASSERT_SUCCESS( fr, &file_system.__file );
	return MISSION_LOGGER_OK;
}

static ml_result_t write_stream( mission_logger_t *sd, const uint8_t *msg, unsigned int size, unsigned int *written )
{
	FRESULT	fr;
	unsigned int i;
	char ascii_str[4];
	uint8_t ascii_size;

	/* Assert there is an open file. */
	ASSERT_OPEN( file_system );

	/* Put the disk io under mutual exclusion. */
	xSemaphoreTake( file_system.__write_mutex, portMAX_DELAY );

	for( i = 0; i < size; ++i )
	{
		ascii_size = byte_to_ascii( *(msg+i), ascii_str );
		fr = f_write( &file_system.__file, ascii_str, ascii_size, written );
		++*written;
#ifdef MISSION_LOGGER_DEBUG
	printf( "f_printf, stream: %d\n", fr );
#endif
	}

	/* Return semaphore so other processes can write / read from file. */
	xSemaphoreGive( file_system.__write_mutex );

	ASSERT_SUCCESS( fr, &file_system.__file );
	return MISSION_LOGGER_OK;
}

static ml_result_t read( mission_logger_t *sd, const char *msg, unsigned int size, unsigned int *bytes_read )
{
	FRESULT fr;

	/* Assert there is an open file to read from. */
	ASSERT_OPEN( file_system );

	/* Put the file io under mutual exclusion. */
	xSemaphoreTake( file_system.__write_mutex, portMAX_DELAY );

	/* Read from the open file. */
	fr = f_read( &file_system.__file, (void *) msg, size, bytes_read );
#ifdef MISSION_LOGGER_DEBUG
	printf( "f_read, read: %d\n", fr );
#endif

	/* Return mutex. */
	xSemaphoreGive( file_system.__write_mutex );

	ASSERT_SUCCESS( fr, &file_system.__file );
	if( *bytes_read != size )
	{
		return MISSION_LOGGER_END_OF_FILE;
	}
	return MISSION_LOGGER_OK;
}

static ml_result_t go_to( mission_logger_t *sd, unsigned int position )
{
	FRESULT fr;

	/* Assert there is an open file to seek in. */
	ASSERT_OPEN( file_system );

	/* Put io under mutual exclusion. */
	xSemaphoreTake( file_system.__write_mutex, portMAX_DELAY );

	/* Seek to desired position. */
	fr = f_lseek( &file_system.__file , position );
#ifdef MISSION_LOGGER_DEBUG
	printf( "f_lseek, go_to: %d\n", fr );
#endif

	/* Return mutex and error check. */
	xSemaphoreGive( file_system.__write_mutex );

	ASSERT_SUCCESS( fr, &file_system.__file );
	return MISSION_LOGGER_OK;
}

static ml_result_t reverse_truncate( mission_logger_t *sd )
{
	FRESULT 		fr;
	UINT			written;
	unsigned int	read;
	ml_result_t		ml_result;
	char 			copy_buffer[COPY_BUFFER_SIZE];

	/* Assert there is an open file to delete from. */
	ASSERT_OPEN( file_system );

	/* Put delete operation under mutual exclusion. */
	xSemaphoreTake( file_system.__delete_mutex, portMAX_DELAY );

	/* Open delete buffer. */
	fr = f_open( &file_system.__delete_buffer, DELETE_BUFFER, FA_READ | FA_WRITE | FA_OPEN_ALWAYS );
#ifdef MISSION_LOGGER_DEBUG
	printf( "f_open, delete buffer: %d\n", fr );
#endif
	if( fr != FR_OK )
	{
		f_close( &file_system.__delete_buffer );
		sd->close( sd );
		xSemaphoreGive( file_system.__delete_mutex );
		return MISSION_LOGGER_FAILURE;
	}

	/* Go to beginning of delete buffer file. */
	fr = f_lseek( &file_system.__delete_buffer, BEGINNING_OF_FILE );
#ifdef MISSION_LOGGER_DEBUG
	printf( "f_lseek, delete buffer: %d\n", fr );
#endif
	if( fr != FR_OK )
	{
		f_close( &file_system.__delete_buffer );
		sd->close( sd );
		xSemaphoreGive( file_system.__delete_mutex );
		return MISSION_LOGGER_FAILURE;
	}

	/* Iterate until all data from current position to end of file is inside */
	/* Delete buffer. */
	ml_result = MISSION_LOGGER_OK;
	while( ml_result != MISSION_LOGGER_END_OF_FILE )
	{
		ml_result = sd->read( sd, copy_buffer, COPY_BUFFER_SIZE, &read );
		fr = f_write( &file_system.__delete_buffer, copy_buffer, read, &written );
#ifdef MISSION_LOGGER_DEBUG
		printf( "f_write, delete buffer: %d\n", fr );
#endif

		if( fr != FR_OK )
		{
			f_close( &file_system.__delete_buffer );
			sd->close( sd );
			xSemaphoreGive( file_system.__delete_mutex );
			return MISSION_LOGGER_FAILURE;
		}
		if(  ml_result == MISSION_LOGGER_FAILURE )
		{
			f_close( &file_system.__delete_buffer );
			sd->close( sd );
			xSemaphoreGive( file_system.__delete_mutex );
			return MISSION_LOGGER_FAILURE;
		}
	}

	/* Move cursor to beginning of file. */
	ml_result = sd->go_to( sd, BEGINNING_OF_FILE );
	if( ml_result != MISSION_LOGGER_OK )
	{
		/* Clean up delete buffer then exit. */
		f_truncate( &file_system.__delete_buffer );
		f_close( &file_system.__delete_buffer );
		sd->close( sd );
		xSemaphoreGive( file_system.__delete_mutex );
		return MISSION_LOGGER_FAILURE;
	}

	/* Truncate file of data. */
	xSemaphoreTake( file_system.__write_mutex, portMAX_DELAY );
	fr = f_truncate( &file_system.__file );
	xSemaphoreGive( file_system.__write_mutex );
#ifdef MISSION_LOGGER_DEBUG
	printf( "f_truncate, delete: %d\n", fr );
#endif
	if( fr != FR_OK )
	{
		f_close( &file_system.__delete_buffer );
		sd->close( sd );
		return MISSION_LOGGER_FAILURE;
	}

	/* Go to beginning of delete buffer file to begin copying data*/
	do {
		fr = f_lseek( &file_system.__delete_buffer, BEGINNING_OF_FILE );
#ifdef MISSION_LOGGER_DEBUG
		printf( "f_lseek, delete buffer: %d\n", fr );
#endif
	} while( fr != FR_OK );

	/* Copy good data from delete buffer back into file. */
	for( ;; )
	{
		/* do while loop prevents data being lost inside delete buffer. */
		/* TODO: better fault tolerance than do while. */
		do {
			fr = f_read( &file_system.__delete_buffer, copy_buffer, COPY_BUFFER_SIZE, &read );
		} while( fr != FR_OK);
		do {
			ml_result = sd->write( sd, copy_buffer, read, &written);
		} while( ml_result != MISSION_LOGGER_OK );
#ifdef MISSION_LOGGER_DEBUG
		printf( "f_read, delete buffer: %d\n", fr );
#endif

		/* All data has been copied back. */
		if( read < COPY_BUFFER_SIZE )
		{
			break;
		}
	}

	/* Go to begining of delete buffer to truncate everything */
	/* TODO: better fault tolerance than loop. */
	do {
		fr = f_lseek( &file_system.__delete_buffer, BEGINNING_OF_FILE );
#ifdef MISSION_LOGGER_DEBUG
		printf( "f_lseek, delete buffer: %d\n", fr );
#endif
	} while( fr != FR_OK );

	/* Truncate delete buffer. */
	/* TODO: better fault tolerance than loop. */
	do {
		fr = f_truncate( &file_system.__delete_buffer );
#ifdef MISSION_LOGGER_DEBUG
		printf( "f_truncate, delete buffer: %d\n", fr );
#endif
	} while( fr != FR_OK );

	/* Close delete buffer. */
	f_close(  &file_system.__delete_buffer );
#ifdef MISSION_LOGGER_DEBUG
	printf( "f_close, close: %d\n", fr );
#endif

	xSemaphoreGive( file_system.__delete_mutex );
	return MISSION_LOGGER_OK;
}

static ml_result_t reformat( mission_logger_t *sd )
{
	/* TODO: write reformat code. */
	return MISSION_LOGGER_FAILURE;
}


/************************************************************************/
/* Private Non Virtual Method											*/
/************************************************************************/
static void fatfs_sd_timer_callback(  TimerHandle_t pxTimer )
{
	/* Do nothing. */
}

static char base_ten[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };
static uint8_t byte_to_ascii( uint8_t byte, char *str )
{
  uint8_t digit, i;
  char temp;

  for(i = 0; i < 3; ++i)
    {
      digit = byte % 10;
      byte = (byte - digit) / 10;
      str[i] = base_ten[digit];
      if( byte == 0 ){ break; }
    }
  if( i == 1 )
    {
      temp = str[0];
      str[0] = str[1];
      str[1] = temp;
    }
  else if( i == 2 )
    {
      temp = str[0];
      str[0] = str[2];
      str[2] = temp;
    }
  return i+1;
}

/************************************************************************/
/* Initialization Method												*/
/************************************************************************/
ml_result_t initialize_mission_logger_lpc_fatfs_sd( mission_logger_t *sd )
{
	sd->open_a = &open_a;
	sd->open_r = &open_r;
	sd->close = &close;
	sd->file_size = &file_size;
	sd->write = &write;
	sd->write_stream = &write_stream;
	sd->read = &read;
	sd->go_to = &go_to;
	sd->delete = &reverse_truncate;
	sd->reformat = &reformat;

	if( file_system.__is_init == ML_INITIALIZED)
	{
		return MISSION_LOGGER_OK;
	}

	file_system.__open_mutex = xSemaphoreCreateMutex();
	file_system.__write_mutex = xSemaphoreCreateMutex();
	file_system.__delete_mutex = xSemaphoreCreateMutex();
	if( file_system.__open_mutex == NULL || file_system.__write_mutex == NULL || file_system.__delete_mutex == NULL )
	{
		return MISSION_LOGGER_FAILURE;
	}
	xSemaphoreGive( file_system.__open_mutex );
	xSemaphoreGive( file_system.__write_mutex );
	xSemaphoreGive( file_system.__delete_mutex );

	sdtimer1 = xTimerCreate("SDTimer1", 100/portTICK_PERIOD_MS, pdFALSE, (void*) 1, fatfs_sd_timer_callback);
	sdtimer2 = xTimerCreate("SDTimer2", 500/portTICK_PERIOD_MS, pdFALSE, (void*) 2, fatfs_sd_timer_callback);

	spi_init( );

#ifdef MISSION_LOGGER_DEBUG
	FRESULT fr = f_mount( &file_system.__fatfs, DEFAULT_DRIVE, MOUNT_LATER);
	printf( "f_mount, init: %d\n", fr );
#else
	f_mount( &file_system.__fatfs, DEFAULT_DRIVE, MOUNT_LATER);
#endif

	file_system.__is_init = ML_INITIALIZED;
	return MISSION_LOGGER_OK;
}
