/*
 * Copyright (C) 2015  Brendan Bruner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * bbruner@ualberta.ca
 *
 * Code here is heavily based on the demo code provided by nxp.
 * http://developer.mbed.org/users/frank26080115/code/LPC1700CMSIS_Examples/file/bf7b9fba3924/I2C/Master_Slave_Interrupt/i2c_master_slave_int_test.c  [March 2015]
 * The only original code is the wrappers and task safety.
 */
/**
 * @file lpc_uart
 * @author Brendan Bruner
 * @date Mar 28, 2015
 */

#include "uart.h"
#include "lpc17xx_uart.h"
#include "lpc_types.h"
#include "LPC17xx.h"
#include "lpc17xx_pinsel.h"
#include "core_cm3.h"
#include "FreeRTOS.h"
#include "semphr.h"
#include "task.h"
#include <stdlib.h>

/********************************************************************/
/* PRIVATE DEFINES													*/
/********************************************************************/
#define UART_NOT_INIT	0
#define UART_IS_INIT 	1

#define UART_GATEWAY_STACK	1024
#define UART_GATEWAY_NAME 	"uart gateway"
#define UART_GATEWAY_PRIO	( tskIDLE_PRIORITY + 4 )
#define UART_GATEWAY_PERIOD	( 100 / portTICK_RATE_MS )

#define UART_BUFFER_SIZE	200

#define UART_RING_BUFSIZE 128
/* Buf mask */
#define __BUF_MASK (UART_RING_BUFSIZE-1)
/* Check buf is full or not */
#define __BUF_IS_FULL(head, tail) ((tail&__BUF_MASK)==((head+1)&__BUF_MASK))
/* Check buf will be full in next receiving or not */
#define __BUF_WILL_FULL(head, tail) ((tail&__BUF_MASK)==((head+2)&__BUF_MASK))
/* Check buf is empty */
#define __BUF_IS_EMPTY(head, tail) ((head&__BUF_MASK)==(tail&__BUF_MASK))
/* Reset buf */
#define __BUF_RESET(bufidx) (bufidx=0)
#define __BUF_INCR(bufidx)  (bufidx=(bufidx+1)&__BUF_MASK)
#define __BUF_SIZE(head, tail) (head-tail)

typedef struct
{
     volatile uint32_t tx_head;                /*!< UART Tx ring buffer head index */
     volatile uint32_t tx_tail;                /*!< UART Tx ring buffer tail index */
     volatile uint32_t rx_head;                /*!< UART Rx ring buffer head index */
     volatile uint32_t rx_tail;                /*!< UART Rx ring buffer tail index */
     volatile uint8_t  tx[UART_RING_BUFSIZE];  /*!< UART Tx data ring buffer */
     volatile uint8_t  rx[UART_RING_BUFSIZE];  /*!< UART Rx data ring buffer */
} UART_RING_BUFFER_T;

/********************************************************************/
/* PRIVATE GLOBAL VARIABLES											*/
/********************************************************************/
static uint8_t is_uart_init;
static SemaphoreHandle_t comm_mutex; /* comm, meaning communication, NOT comm board. */

/********************************************************************/
/* PUBLIC GLOBAL VARIABLES											*/
/********************************************************************/
UART_RING_BUFFER_T ring_buffer;
volatile FlagStatus TxIntStat;

/********************************************************************/
/* PRIVATE METHOD DELCARATIONS										*/
/********************************************************************/
static void pinselect_uart2();
static void init_uart2();
static void uart2_fifo_config();
static void UART_IntErr(uint8_t bLSErrType);
static void uart_irq_transmit_handle(void);
static void uart_irq_receive_handle(void);

/********************************************************************/
/* PRIVATE METHOD DEFINES											*/
/********************************************************************/
void UART2_IRQHandler(void)
{
    uint32_t intsrc, tmp, tmp1;

    /* Determine the interrupt source */
    intsrc = UART_GetIntId(LPC_UART2);
    tmp = intsrc & UART_IIR_INTID_MASK;

    // Receive Line Status
    if (tmp == UART_IIR_INTID_RLS){
        // Check line status
        tmp1 = UART_GetLineStatus(LPC_UART2);
        // Mask out the Receive Ready and Transmit Holding empty status
        tmp1 &= (UART_LSR_OE | UART_LSR_PE | UART_LSR_FE \
                | UART_LSR_BI | UART_LSR_RXFE);
        // If any error exist
        if (tmp1) {
                UART_IntErr(tmp1);
        }
    }

    // Receive Data Available or Character time-out
    if ((tmp == UART_IIR_INTID_RDA) || (tmp == UART_IIR_INTID_CTI)){
            uart_irq_receive_handle();
    }

    // Transmit Holding Empty
    if (tmp == UART_IIR_INTID_THRE){
            uart_irq_transmit_handle();
    }

}

static void uart_irq_receive_handle(void)
{
    uint8_t tmpc;
    uint32_t rLen;

    while(1){
        // Call UART read function in UART driver
        rLen = UART_Receive((LPC_UART_TypeDef *)LPC_UART2, &tmpc, 1, NONE_BLOCKING);
        // If data received
        if (rLen){
            /* Check if buffer is more space
             * If no more space, remaining character will be trimmed out
             */
            if (!__BUF_IS_FULL(ring_buffer.rx_head,ring_buffer.rx_tail)){
                ring_buffer.rx[ring_buffer.rx_head] = tmpc;
                __BUF_INCR(ring_buffer.rx_head);
            }
        }
        // no more data
        else {
            break;
        }
    }
}

static void uart_irq_transmit_handle(void)
{
    // Disable THRE interrupt
    UART_IntConfig((LPC_UART_TypeDef *)LPC_UART2, UART_INTCFG_THRE, DISABLE);

    /* Wait for FIFO buffer empty, transfer UART_TX_FIFO_SIZE bytes
     * of data or break whenever ring buffers are empty */
    /* Wait until THR empty */
    while (UART_CheckBusy((LPC_UART_TypeDef *)LPC_UART2) == SET);

    while (!__BUF_IS_EMPTY(ring_buffer.tx_head,ring_buffer.tx_tail))
    {
        /* Move a piece of data into the transmit FIFO */
        if (UART_Send((LPC_UART_TypeDef *)LPC_UART2, (uint8_t *)&ring_buffer.tx[ring_buffer.tx_tail], 1, NONE_BLOCKING)){
        /* Update transmit ring FIFO tail pointer */
        __BUF_INCR(ring_buffer.tx_tail);
        } else {
            break;
        }
    }

    /* If there is no more data to send, disable the transmit
       interrupt - else enable it or keep it enabled */
    if (__BUF_IS_EMPTY(ring_buffer.tx_head, ring_buffer.tx_tail)) {
        UART_IntConfig((LPC_UART_TypeDef *)LPC_UART2, UART_INTCFG_THRE, DISABLE);
        // Reset Tx Interrupt state
        TxIntStat = RESET;
    }
    else{
        // Set Tx Interrupt state
        TxIntStat = SET;
        UART_IntConfig((LPC_UART_TypeDef *)LPC_UART2, UART_INTCFG_THRE, ENABLE);
    }
}

void UART_IntErr(uint8_t bLSErrType)
{
//  uint8_t test;
//	volatile FlagStatus exitflag;
//	uint32_t idx, len;
//	uint8_t buffer[10];
	pinselect_uart2();
	init_uart2();
	uart2_fifo_config();
	UART_TxCmd(LPC_UART2, ENABLE);
	/* Enable UART Rx interrupt */
	UART_IntConfig((LPC_UART_TypeDef *)LPC_UART2, UART_INTCFG_RBR, ENABLE);
	/* Enable UART line status interrupt */
	UART_IntConfig((LPC_UART_TypeDef *)LPC_UART2, UART_INTCFG_RLS, ENABLE);
	TxIntStat = RESET;
	__BUF_RESET(ring_buffer.rx_head);
	__BUF_RESET(ring_buffer.rx_tail);
	__BUF_RESET(ring_buffer.tx_head);
	__BUF_RESET(ring_buffer.tx_tail);
	/* preemption = 1, sub-priority = 1 */
	NVIC_SetPriority(UART2_IRQn, ((0x01<<3)|0x01));
	/* Enable Interrupt for UART0 channel */
	NVIC_EnableIRQ(UART2_IRQn);
	// reset exit flag
//	exitflag = RESET;
}

void pinselect_uart2()
{
	PINSEL_CFG_Type PinCfg;
	PinCfg.Funcnum = PINSEL_FUNC_1;
	PinCfg.OpenDrain = PINSEL_PINMODE_NORMAL;
	PinCfg.Portnum = PINSEL_PORT_0;
	PinCfg.Pinnum = PINSEL_PIN_10;
	PINSEL_ConfigPin(&PinCfg); //transmitter
	PinCfg.Pinnum = PINSEL_PIN_11;
	PINSEL_ConfigPin(&PinCfg); //receiver
}

void init_uart2()
{
	UART_CFG_Type uartCfg;
	uartCfg.Baud_rate = 115200;
	uartCfg.Databits = UART_DATABIT_8;
	uartCfg.Parity = UART_PARITY_NONE;
	uartCfg.Stopbits = UART_STOPBIT_1;
	UART_Init(LPC_UART2, &uartCfg);
}

void uart2_fifo_config()
{
	UART_FIFO_CFG_Type FIFOCfg;
	UART_FIFOConfigStructInit(&FIFOCfg);
	FIFOCfg.FIFO_Level = UART_FIFO_TRGLEV0;
	UART_FIFOConfig(LPC_UART2, &FIFOCfg);
}

/********************************************************************/
/* PUBLIC METHOD DEFINES											*/
/********************************************************************/
void serial2_begin( void )
{
	if( is_uart_init == UART_IS_INIT )
	{
		return;
	}
	comm_mutex = xSemaphoreCreateMutex();
	is_uart_init = UART_IS_INIT;
//	volatile FlagStatus exitflag;
//	uint32_t idx, len;
//	uint8_t buffer[10];
	pinselect_uart2();
	init_uart2();
	uart2_fifo_config();
	UART_TxCmd(LPC_UART2, ENABLE);
	/* Enable UART Rx interrupt */
	UART_IntConfig((LPC_UART_TypeDef *)LPC_UART2, UART_INTCFG_RBR, ENABLE);
	/* Enable UART line status interrupt */
	UART_IntConfig((LPC_UART_TypeDef *)LPC_UART2, UART_INTCFG_RLS, ENABLE);
	TxIntStat = RESET;
	__BUF_RESET(ring_buffer.rx_head);
	__BUF_RESET(ring_buffer.rx_tail);
	__BUF_RESET(ring_buffer.tx_head);
	__BUF_RESET(ring_buffer.tx_tail);
	/* preemption = 1, sub-priority = 1 */
	NVIC_SetPriority(UART2_IRQn, ((0x01<<3)|0x01));
	/* Enable Interrupt for UART0 channel */
	NVIC_EnableIRQ(UART2_IRQn);
	// reset exit flag
//	exitflag = RESET;
	//xTaskCreate( uart_gateway_transmit, UART_GATEWAY_NAME, UART_GATEWAY_STACK, NULL, UART_GATEWAY_PRIO, NULL );
	//xTaskCreate( uart_gateway_receive, UART_GATEWAY_NAME, UART_GATEWAY_STACK, NULL, UART_GATEWAY_PRIO, NULL );

}

uint32_t serial2_send( uint8_t const *data, uint8_t buflen )
{
	uint32_t bytes = 0;

	/* Temporarily lock out UART transmit interrupts during this
	read so the UART transmit interrupt won't cause problems
	with the index values */
	UART_IntConfig(LPC_UART2, UART_INTCFG_THRE, DISABLE);

	/* Loop until transmit run buffer is full or until n_bytes
	expires */
	while ((buflen > 0) && (!__BUF_IS_FULL(ring_buffer.tx_head, ring_buffer.tx_tail)))
	{
		/* Write data from buffer into ring buffer */
		ring_buffer.tx[ring_buffer.tx_head] = *data;
		data++;

		/* Increment head pointer */
		__BUF_INCR(ring_buffer.tx_head);

		/* Increment data count and decrement buffer size count */
		bytes++;
		buflen--;
	}

	/*
	 * Check if current Tx interrupt enable is reset,
	 * that means the Tx interrupt must be re-enabled
	 * due to call UART_IntTransmit() function to trigger
	 * this interrupt type
	 */
	if (TxIntStat == RESET) {
		uart_irq_transmit_handle();
	}
	/*
	* Otherwise, re-enables Tx Interrupt
	*/
	else {
		UART_IntConfig(LPC_UART2, UART_INTCFG_THRE, ENABLE);
	}

	return bytes;
}
uint32_t serial2_read( uint8_t *data, uint8_t buflen )
{
    uint32_t bytes = 0;

    /* Temporarily lock out UART receive interrupts during this
       read so the UART receive interrupt won't cause problems
       with the index values */
    UART_IntConfig(LPC_UART2, UART_INTCFG_RBR, DISABLE);

    /* Loop until receive buffer ring is empty or
        until max_bytes expires */
    while ((buflen > 0) && (!(__BUF_IS_EMPTY(ring_buffer.rx_head, ring_buffer.rx_tail))))
    {
        /* Read data from ring buffer into user buffer */
        *data = ring_buffer.rx[ring_buffer.rx_tail];
        data++;

        /* Update tail pointer */
        __BUF_INCR(ring_buffer.rx_tail);

        /* Increment data count and decrement buffer size count */
        bytes++;
        buflen--;
    }

    /* Re-enable UART interrupts */
    UART_IntConfig(LPC_UART2, UART_INTCFG_RBR, ENABLE);

    return bytes;
}

uint32_t serial2_available( void )
{
	return __BUF_SIZE(ring_buffer.rx_head, ring_buffer.rx_tail);
}

void serial2_lock( void )
{
	xSemaphoreTake( comm_mutex, portMAX_DELAY );
}
void serial2_unlock( void )
{
	xSemaphoreGive( comm_mutex );
}

void serial2_reset( void )
{
	__BUF_RESET(ring_buffer.rx_head);
	__BUF_RESET(ring_buffer.rx_tail);
	__BUF_RESET(ring_buffer.tx_head);
	__BUF_RESET(ring_buffer.tx_tail);
}
