#include <stdlib.h>
#include <stdio.h>
#include "spi_sd_lpc17xx.h"
#include "ff.h"
#include "stdlib.h"
#include "FreeRTOS.h"
#include "task.h"
#include "system_states.h"
#include "system_state_relay.h"
#include <stdint.h>
#include "commands.h"
#include "driver_toolkit.h"
#include "mission_logger.h"
#include "lpc17xx_uart.h"
#include "lpc17xx_pinsel.h"
#include "core_cm3.h"
#include "lpc17xx_timer.h"
#include "i2c.h"
#include "mission_logger.h"
#include "uart.h"
#include "task_defines.h"


/****************************************************************************/
/* TASK CONTROL																*/
/****************************************************************************/
//#define USE_SD_EXAMPLE
#define USE_HK_TASK
#define USE_TELECOMMAND_TASK
#define USE_TRANSMIT_TASK
//#define USE_PRINT_TIME
//#define USE_I2C_EXAMPLE
//#define USE_UART_EXAMPLE


/****************************************************************************/
/* DEFINES																	*/
/****************************************************************************/


/****************************************************************************/
/* FUNCTION PROTOTYPES														*/
/****************************************************************************/
void vTimerCallback1( TimerHandle_t pxTimer );
void vTimerCallback2( TimerHandle_t pxTimer );
static void state_task( void *pvParameters );
static void sd_task( void *pvParameters );
static void test_send_time_task(void *pvParameters);
static void test_recv_time_task(void *pvParameters);
static void hk_task( void * );
static void init_task( void * );
static void telecommand_task( void * );
static void transmit_task( void * );
static void sd_task( void * );
static void i2c_ex( void * );
static void uart_test( void * );
static void print_time( void * );


/****************************************************************************/
/* STATIC VARIABLES															*/
/****************************************************************************/
#ifdef USE_SD_EXAMPLE
static mission_logger_t logger;
#endif

static command_t 				collect_eps_hk, collect_adcs_hk, collect_comm_hk;
static system_state_relay_t 	relay;
static driver_toolkit_t			drivers;

/****************************************************************************/
/* MAIN																		*/
/****************************************************************************/
int main(void)
{
	printf( "Starting up...\n" );
#ifdef USE_I2C_EXAMPLE
	setupI2C( );
#endif
#ifdef USE_UART_EXAMPLE
	serial2_begin();
#endif
	xTaskCreate( init_task, INIT_NAME, INIT_STACK, NO_PARAMETERS, INIT_PRIO, NO_HANDLE);
	vTaskStartScheduler();
	return 0;
}


/****************************************************************************/
/* FUNCTION IMPLEMENTATIONS													*/
/****************************************************************************/
static void init_task( void * args )
{
	setupI2C( );

	/* Initialize driver toolkit with the pre initialized drivers */
	initialize_driver_toolkit_lpc_local( &drivers );
	initialize_comm_arduino_remote( &drivers.comm );
	initialize_adcs_arduino_remote( &drivers.adcs );

	/* Initialize state relay */
	initialize_system_state_relay	( &relay, NULL, &drivers );

	/* Initialize commands being used */
	initialize_command_collect_eps_hk( &collect_eps_hk, &drivers );
	initialize_command_collect_adcs_hk( &collect_adcs_hk, &drivers );
	initialize_command_collect_comm_hk( &collect_comm_hk, &drivers );

#ifdef USE_TRANSMIT_TASK
	xTaskCreate( transmit_task, TRANSMIT_NAME, TRANSMIT_STACK, NO_PARAMETERS, TRANSMIT_PRIO, NO_HANDLE );
#endif
#ifdef USE_HK_TASK
	xTaskCreate( hk_task, HK_NAME, HK_STACK, NO_PARAMETERS, HK_PRIO, NO_HANDLE );
#endif
#ifdef USE_TELECOMMAND_TASK
	xTaskCreate( telecommand_task, TELECOMMAND_NAME, TELECOMMAND_STACK, NO_PARAMETERS, TELECOMMAND_PRIO, NO_HANDLE );
#endif
#ifdef USE_SD_EXAMPLE
	xTaskCreate( sd_task, SD_NAME, SD_STACK, NO_PARAMETERS, SD_PRIO, NO_HANDLE );
#endif
#ifdef USE_I2C_EXAMPLE
	xTaskCreate( i2c_ex, I2C_NAME, I2C_STACK, NO_PARAMETERS, I2C_PRIO, NO_HANDLE );
#endif
#ifdef USE_UART_EXAMPLE
	xTaskCreate( uart_test, UART_NAME, UART_STACK, NO_PARAMETERS, UART_PRIO, NO_HANDLE );
#endif
#ifdef USE_PRINT_TIME
	xTaskCreate( print_time, PRINT_NAME, PRINT_STACK, NO_PARAMETERS, PRINT_PRIO, NO_HANDLE );
#endif
	vTaskDelete( NULL );
	for( ;; ) /* For loop is a fail safe. */
	{
		vTaskDelay( DELAY_ONE_SECOND );
	}
}

#ifdef USE_PRINT_TIME
static void print_time( void *args )
{
	for( ;; )
	{
		printf( " \b" );
		vTaskDelay( PRINT_DELAY );
	}
}
#endif

#ifdef USE_I2C_EXAMPLE
static void i2c_ex( void *args )
{
	lcd_init( );
	for( ;; )
	{
		lcd_seek( 0 );
		lcd_write("hello", 5);
		vTaskDelay( ONE_SECOND );
		lcd_seek(8);
		lcd_write("derp", 4);
		vTaskDelay( ONE_SECOND);
	}
}
#endif

#ifdef USE_UART_EXAMPLE
void uart_test( void *args )
{
	comm_t comm;
	nanocom_hk_t hk[HK_ARRAY_SIZE];
	uint8_t retreived;
	int i;
	uint8_t msg[40];
	for( i = 0; i < 40; ++i )
	{
		msg[i] = i;
		printf( "msg[%d] = %d, ", i, i );
	}

	printf( "\n" );
	initialize_comm_arduino_remote( &comm );
	for( ;; )
	{
		retreived = comm.transmit( &comm, msg, 40 );
		printf( "transmitted: %d\n", retreived );

		vTaskDelay( UART_SLEEP_PERIOD );
	}
}
#endif

#ifdef USE_HK_TASK
static void hk_task( void *args )
{
	for( ;; )
	{
		relay_command_collect_housekeeping( &relay, &collect_eps_hk );
		relay_command_collect_housekeeping( &relay, &collect_adcs_hk );
		relay_command_collect_housekeeping( &relay, &collect_comm_hk );

		vTaskDelay( HK_SLEEP_PERIOD );
	}
}
#endif

#ifdef USE_TELECOMMAND_TASK
static void telecommand_task( void *arg )
{
	uint8_t telecommand_confirmation;
	telecommand_packet_t packet;
	command_t telecommand;
	comm_t *comm;

	comm = &drivers.comm;
	for( ;; )
	{
		telecommand_confirmation = comm->poll_telecommand( comm, &packet );
		if( telecommand_confirmation == INCOMING_TELECOMMAND )
		{
			if( extract_code( &packet ) == 0 )
			{
				switch extract_priority( &packet )
				{
				case 0:
					initialize_command_print_string( &telecommand, "command 0, priority 0\n" );
					break;
				case 1:
					initialize_command_print_string( &telecommand, "command 0, priority 1\n" );
					break;
				case 2:
					initialize_command_print_string( &telecommand, "command 0, priority 2\n" );
					break;
				case 3:
					initialize_command_print_string( &telecommand, "command 0, priority 3\n" );
					break;
				case 4:
					initialize_command_print_string( &telecommand, "command 0, priority 4\n" );
					break;
				}
				relay_command_very_high_priority( &relay, &telecommand );
			}
			else if( extract_code( &packet ) == 256 )
			{
				initialize_command_print_state( &telecommand, &relay );
				relay_command_very_high_priority( &relay, &telecommand );
			}
		}

		vTaskDelay( TELECOMMAND_SLEEP_PERIOD );
	}
}
#endif

#ifdef USE_TRANSMIT_TASK
static void transmit_task( void *arg )
{
	command_t transmit_eps, transmit_adcs, transmit_comm, transmit_state;

	initialize_command_transmit_eps_hk( &transmit_eps, &drivers );
	initialize_command_transmit_adcs_hk( &transmit_adcs, &drivers );
	initialize_command_transmit_comm_hk( &transmit_comm, &drivers );
	initialize_command_transmit_state_hk( &transmit_state, &drivers );
	for( ;; )
	{
		relay_command_transmit_data( &relay, &transmit_comm );
		relay_command_transmit_data( &relay, &transmit_eps );
		relay_command_transmit_data( &relay, &transmit_adcs );
		relay_command_transmit_data( &relay, &transmit_state );
		vTaskDelay( TRANSMIT_SLEEP_PERIOD );
	}
}
#endif

static void test_send_time_task(void *pvParameters)
{
	command_t 				send_time;
	system_state_relay_t 	relay;
	driver_toolkit_t		drivers;
	int i;

	/* Initialize driver toolkit with the pre initialized drivers */
	initialize_driver_toolkit_lpc_local( &drivers );

	/* Initialize state relay */
	initialize_system_state_relay	( &relay, NULL, &drivers );

	/* Initialize commands being used */
	initialize_command_send_time_to_adcs_over_i2c( &send_time, &drivers );
	printf("time sent \n");


	for( ;; )
	{
		relay_command_very_high_priority( &relay, &send_time );
		vTaskDelay( 2000 / portTICK_RATE_MS );
	}
}

static void test_recv_time_task(void *pvParameters)
{
	command_t 				recv_time;
	system_state_relay_t 	relay;
	driver_toolkit_t		drivers;
	int i;

	/* Initialize driver toolkit with the pre initialized drivers */
	initialize_driver_toolkit_lpc_local( &drivers );

	/* Initialize state relay */
	initialize_system_state_relay	( &relay, NULL, &drivers );

	/* Initialize commands being used */
	initialize_command_receive_time_from_adcs_over_i2c( &recv_time, &drivers );
	printf("time received \n");


	for( ;; )
	{
		relay_command_very_high_priority( &relay, &recv_time );
		vTaskDelay( 2000 / portTICK_RATE_MS );
	}
}

#ifdef USE_SD_EXAMPLE
static void sd_task( void *arg )
{
	unsigned int written, read;
	unsigned long size;
	char data[5];
	int count = 0;
	uint8_t test[] = {'d','e','r','p',10,0,'d','e','r','p'};

	data[4] = '\0';
	initialize_mission_logger_lpc_fatfs_sd( &logger );

	for( ;; )
	{
		logger.open_a( &logger, "test.txt" );
		logger.write( &logger, (char const *) test, 10, &written );
		logger.close( &logger );
		vTaskDelay( DELAY_ONE_SECOND );
		printf("done\n");
	}

	for( count = 0; count < 20; ++count )
	{
		logger.open_a( &logger, "delete_file.txt" );
		logger.write( &logger, "delete", 6, &written );
		logger.close( &logger );

		logger.open_a( &logger, "no_delete_file.txt" );
		logger.write( &logger, "delete", 6, &written );
		logger.close( &logger );
	}

	logger.open_a( &logger, "delete_file.txt" );
	logger.go_to( &logger, 9 );
	logger.delete( &logger );
	logger.close( &logger );

	for(;;)
	{
		printf( "done\n" );
		vTaskDelay( 1000 / portTICK_RATE_MS );
	}
}
#endif

void vApplicationMallocFailedHook( void )
{
	/* This function will only be called if an API call to create a task, queue
	or semaphore fails because there is too little heap RAM remaining. */
	for( ;; );
}
/*-----------------------------------------------------------*/

void vApplicationStackOverflowHook( xTaskHandle *pxTask, signed char *pcTaskName )
{
	/* This function will only be called if a task overflows its stack.  Note
	that stack overflow checking does slow down the context switch
	implementation. */
	for( ;; );
}

void vTimerCallback( TimerHandle_t pxTimer )
 {
    /* Empty callback function for timer. */
 }

